"""
Font Info
"""
import logging
import re

import wx
from fontTools.ufoLib import fontInfoAttributesVersion3ValueData
from wbDefcon import Font, Info
from wbBase.scripting import MacroButtonMixin

from ...control.libControl import FontLibControl, LibControlValidator
from .fontInfoCaret import FontInfoCaret
from .fontInfoCodepage import FontInfoCodepage
from .fontInfoControl import BitList
from .fontInfoClassification import FontInfoClassification
from .fontInfoEmbedding import FontInfoEmbedding
from .fontInfoEncoding import FontInfoEncoding
from .fontInfoHinting import FontInfoHinting
from .fontInfoHintingPS import FontInfoHintingPS
from .fontInfoIdentification import FontInfoIdentification
from .fontInfoLegal import FontInfoLegal
from .fontInfoLine import FontInfoLine
from .fontInfoMetric import FontInfoMetric
from .fontInfoName import FontInfoName
from .fontInfoNote import FontInfoNote
from .fontInfoSuperSubscript import FontInfoSuperSubscript
from .fontInfoUnicode import FontInfoUnicode
from .fontInfoValidator import (
    FontInfoBitListValidator,
    FontInfoCheckBoxValidator,
    FontInfoIntlistTextCtrlValidator,
    FontInfoTextCtrlValidator,
    FontInfoPShintCtrlValidator,
)
from .fontInfoNaviagtionPanelUI import FontInfoNaviagtionPanelUI
from .hintPSpanel import ZonePanel, StemPanel

log = logging.getLogger(__name__)

FontInfoAttributes = Info._properties.keys()
# pattern for Bit-List FontInfoAttributes
bitPattern = re.compile(r"^\w+ \d$")


class InfoTreeCtrl(wx.TreeCtrl):
    """
    Tree control with FontInfo
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE
        | wx.TR_HIDE_ROOT
        | wx.TR_LINES_AT_ROOT
        | wx.TR_HAS_BUTTONS
        | wx.TR_TWIST_BUTTONS
        | wx.TR_NO_LINES
        | wx.TR_FULL_ROW_HIGHLIGHT,
    ):
        super().__init__(parent, id, pos, size, style, name="InfoTreeCtrl")
        self._font = None
        self.root = self.AddRoot("FontInfo")
        itemID = self.AppendItem(self.root, "Names", data=0)
        identID = self.AppendItem(self.root, "Identification", data=1)
        itemID = self.AppendItem(identID, "Classification", data=2)
        legalID = self.AppendItem(self.root, "Legal", data=3)
        itemID = self.AppendItem(legalID, "Embedding", data=4)
        metricID = self.AppendItem(self.root, "Metric", data=5)
        itemID = self.AppendItem(metricID, "Super- and Subscript", data=6)
        itemID = self.AppendItem(metricID, "Line", data=7)
        itemID = self.AppendItem(metricID, "Caret", data=8)
        encodingID = self.AppendItem(self.root, "Encoding", data=9)
        itemID = self.AppendItem(encodingID, "Codepage", data=10)
        itemID = self.AppendItem(encodingID, "Unicode", data=11)
        hintingID = self.AppendItem(self.root, "Hinting", data=12)
        itemID = self.AppendItem(hintingID, "PostScript Hinting", data=13)
        itemID = self.AppendItem(self.root, "Note", data=14)
        itemID = self.AppendItem(self.root, "Lib", data=15)
        self.Bind(wx.EVT_TREE_SEL_CHANGED, self.on_selectionChanged)
        self.Bind(wx.EVT_TREE_SEL_CHANGING, self.on_selectionChanging)

    @property
    def infoBookCtrl(self):
        return self.GrandParent.infoBookCtrl

    def on_selectionChanged(self, event):
        if self:
            pageNum = self.GetItemData(event.Item)
            self.infoBookCtrl.ChangeSelection(pageNum)
            self.infoBookCtrl.CurrentPage.TransferDataToWindow()
            # self.infoBookCtrl.CurrentPage.Validator.TransferToWindow()
        event.Skip()

    def on_selectionChanging(self, event):
        if self:
            pageNum = self.GetItemData(event.Item)
            log.debug("on_selectionChanging %r", pageNum)
            if pageNum is not None:
                page = self.infoBookCtrl.GetPage(pageNum)
                if not page.Validate():
                    log.debug("on_selectionChanging %r.Validate() failed", page)
                    event.Veto()
                    return
        event.Skip()


class InfoBookCtrl(wx.Simplebook):
    """
    Book control with FontInfo pages
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE,
    ):
        super().__init__(parent, id, pos, size, style, name="InfoBookCtrl")
        self.SetExtraStyle(wx.WS_EX_VALIDATE_RECURSIVELY)
        self._font = None
        self.fontInfoControls = {}
        pages = (
            ("Name", FontInfoName),
            ("Identification", FontInfoIdentification),
            ("Classification", FontInfoClassification),
            ("Legal", FontInfoLegal),
            ("Embedding", FontInfoEmbedding),
            ("Metric", FontInfoMetric),
            ("Super- and Subscript", FontInfoSuperSubscript),
            ("Line", FontInfoLine),
            ("Caret", FontInfoCaret),
            ("Encoding", FontInfoEncoding),
            ("Codepage", FontInfoCodepage),
            ("Unicode", FontInfoUnicode),
            ("Hinting", FontInfoHinting),
            ("HintingPS", FontInfoHintingPS),
            ("Note", FontInfoNote),
            ("Lib", FontLibControl),
        )
        self.Freeze()
        for name, page in pages:
            self.AddPage(page(self), name)

        self._collectControls(self)
        self._setControlAttributes()
        self._bindControlEvents()
        self.ChangeSelection(0)
        self.Thaw()
        self.Bind(wx.EVT_BOOKCTRL_PAGE_CHANGING, self.on_pageChanging)
        self.Bind(wx.EVT_BOOKCTRL_PAGE_CHANGED, self.on_pageChanged)

    def __repr__(self):
        return f"<{self.__class__.__name__} of {self.Parent}>"

    @property
    def font(self):
        """The wbDefcon Font object of this InfoPage"""
        return self._font

    @font.setter
    def font(self, value):
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.GrandParent._font
        self._font = value
        for child in self.Children:
            child.TransferDataToWindow()
        self._font.info.addObserver(self, "handleNotification", "Info.ValueChanged")
        log.debug("Font %r set for %r", self._font, self)
        self.CurrentPage.TransferDataToWindow()

    @font.deleter
    def font(self):
        self._font.info.removeObserver(self, "Info.ValueChanged")
        self._font = None

    def getFontInfo(self):
        if self._font is not None:
            return self._font.info

    def getFontLib(self):
        if self._font is not None:
            return self._font.lib

    def _collectControls(self, win):
        if win.Children:
            for child in win.Children:
                if isinstance(child, FontLibControl):
                    child.Validator = LibControlValidator(self.getFontLib)
                if child.Name in FontInfoAttributes:
                    self.fontInfoControls[child.Name] = child
                elif (
                    bitPattern.match(child.Name)
                    and child.Name.split(" ", 1)[0] in FontInfoAttributes
                ):
                    self.fontInfoControls[child.Name] = child
                self._collectControls(child)

    def _getControlValueType(self, controlName):
        if controlName in fontInfoAttributesVersion3ValueData:
            return fontInfoAttributesVersion3ValueData[controlName]["type"]

    def _setControlAttributes(self):
        for name, control in self.fontInfoControls.items():
            control.ToolTip = f"info.{name}"
            valueType = self._getControlValueType(name)
            if isinstance(control, wx.TextCtrl):
                if valueType == "integerList":
                    control.Validator = FontInfoIntlistTextCtrlValidator(
                        self.getFontInfo
                    )
                else:
                    control.Validator = FontInfoTextCtrlValidator(self.getFontInfo)
                    if valueType in (int, float, (int, float), (float, int)):
                        control.WindowStyle |= wx.TE_RIGHT
                    control.WindowStyle |= wx.TE_NOHIDESEL
                    if control.IsSingleLine():
                        control.SetHint("None")
            elif isinstance(control, wx.CheckBox):
                control.Validator = FontInfoCheckBoxValidator(self.getFontInfo)
            elif isinstance(control, BitList):
                control.Validator = FontInfoBitListValidator(self.getFontInfo)
            elif isinstance(control, (ZonePanel, StemPanel)):
                control.Validator = FontInfoPShintCtrlValidator(self.getFontInfo)

    def _bindControlEvents(self):
        for control in self.fontInfoControls.values():
            control.Bind(wx.EVT_KILL_FOCUS, self.on_controlKillFocus)

    def handleNotification(self, notification):
        if notification.name == "Info.ValueChanged":
            log.debug("InfoBookCtrl.handleNotification(%r)", notification.data)
            attribute = notification.data["attribute"]
            if attribute in ("openTypeOS2Selection",):
                for controlName, control in self.fontInfoControls.items():
                    if controlName.startswith(attribute + " "):
                        control.Validator.TransferToWindow()
            else:
                control = self.fontInfoControls.get(attribute)
                if control and control.Validator:
                    control.Validator.TransferToWindow()

    # -----------------------------------------------------------------------------
    # Event handler
    # -----------------------------------------------------------------------------

    def on_pageChanging(self, event):
        page = self.GetPage(event.OldSelection)
        if page.Validate():
            if page.TransferDataFromWindow():
                event.Skip()
            else:
                log.debug("TransferDataFromWindow failed for %r", page)
                event.Veto()
        else:
            log.debug("page Validate failed for %r", page)
            event.Veto()

    def on_pageChanged(self, event):
        event.Skip()

    def on_controlKillFocus(self, event):
        log.debug("on_controlKillFocus %r", event.EventObject.Name)
        control = event.EventObject
        if control.Validator and control.Validator.Validate(control.Parent):
            log.debug("%r valid", event.EventObject.Name)
            control.Validator.TransferFromWindow()
        else:
            control.SelectAll()
            wx.CallAfter(control.SetFocus)
        event.Skip()


class FontInfoNaviagtionPanel(FontInfoNaviagtionPanelUI, MacroButtonMixin):
    """
    Naviagtion Panel of InfoPage
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.TAB_TRAVERSAL,
        name="FontInfoNaviagtionPanel",
    ):
        FontInfoNaviagtionPanelUI.__init__(
            self, parent, id=id, pos=pos, size=size, style=style, name=name
        )
        MacroButtonMixin.__init__(
            self, self.button_macro, "_fontInfo", self.GrandParent.view
        )

    # =========================================================================
    # Properties
    # =========================================================================

    @property
    def font(self):
        return self.Parent.font

    @property
    def currentPage(self):
        """Active page of InfoBookCtrl"""
        return self.Parent.infoBookCtrl.CurrentPage

    @property
    def currentAttributes(self):
        attributes = []

        def _collectAttributes(control):
            if control.Children:
                for child in control.Children:
                    if child.Name in FontInfoAttributes:
                        attributes.append(child.Name)
                    elif bitPattern.match(child.Name):
                        childName = child.Name.split(" ", 1)[0]
                        if (
                            childName in FontInfoAttributes
                            and childName not in attributes
                        ):
                            attributes.append(childName)
                    _collectAttributes(child)

        _collectAttributes(self.currentPage)
        return attributes

    @property
    def currentInfo(self):
        info = {}
        fontInfo = self.font.info
        for attr in self.currentAttributes:
            info[attr] = getattr(fontInfo, attr)
        return info

    # =========================================================================
    # Event Handler
    # =========================================================================

    def on_button_copy(self, event):
        from ... import AllFonts, CurrentFont, SelectFonts

        info = self.currentInfo
        name = self.currentPage.Name
        for font in SelectFonts(
            f"Select fonts to copy {name} to:",
            f"Copy {name}",
            [f for f in AllFonts() if f is not CurrentFont()],
        ):
            fontInfo = font.info.naked()
            for attrName, value in info.items():
                setattr(fontInfo, attrName, value)

    def onUpdate_button_copy(self, event):
        invalidNames = ("splitterWindow",)
        fontCount = len(
            [
                d
                for d in wx.GetApp().documentManager.documents
                if d.typeName == "UFO document"
            ]
        )
        event.Enable(self.currentPage.Name not in invalidNames and fontCount > 1)


class InfoPage(wx.Panel):
    """
    Main Font InfoPage with
        - InfoTreeCtrl
        - InfoBookCtrl
        - FontInfoNaviagtionPanel
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE | wx.TAB_TRAVERSAL,
    ):
        super().__init__(parent, id, pos, size, style, name="InfoPage")
        self.SetExtraStyle(wx.WS_EX_VALIDATE_RECURSIVELY)
        self._font = None
        self.Freeze()
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.infoPanel = wx.SplitterWindow(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SP_3D | wx.SP_NOBORDER | wx.BORDER_NONE,
        )
        self.infoPanel.SetExtraStyle(wx.WS_EX_VALIDATE_RECURSIVELY)
        self.infoTreeCtrl = InfoTreeCtrl(self.infoPanel)
        self.infoBookCtrl = InfoBookCtrl(self.infoPanel)

        sizer.Add(self.infoPanel, 1, wx.EXPAND, 0)
        self.infoPanel.SetMinimumPaneSize(20)
        self.infoPanel.SashGravity = 0.3
        self.infoPanel.SplitVertically(self.infoTreeCtrl, self.infoBookCtrl, 200)

        self.infoStatusPanel = FontInfoNaviagtionPanel(self)
        sizer.Add(self.infoStatusPanel, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        self.Layout()
        self.Thaw()

        # Connect Events
        self.infoPanel.Bind(wx.EVT_IDLE, self.infoPanelOnIdle)

    def __repr__(self):
        return f"<{self.__class__.__name__} of {self.Parent}>"

    @property
    def font(self):
        """The wbDefcon Font object of this InfoPage"""
        return self._font

    @font.setter
    def font(self, value):
        if value == self._font:
            return
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.Parent.font
        self._font = value
        log.debug("Font %r set for %r", self._font, self)
        self.infoBookCtrl.font = value

    @font.deleter
    def font(self):
        del self.infoBookCtrl.font
        self._font = None

    def infoPanelOnIdle(self, event):
        self.infoPanel.SetSashPosition(200)
        self.infoPanel.Unbind(wx.EVT_IDLE)

    def Destroy(self):
        del self.font
        return super().Destroy()

    # def TransferDataFromWindow(self):
    #     value = super().TransferDataFromWindow()
    #     log.debug("TransferDataFromWindow for %r returns %r", self, value)
    #     return value



"""
componentDialog
===============================================================================
"""

import wx

from .componentDialogUI import ComponentDialogUI
from .findGlyphDialog import FindGlyphDialog


class ComponentDialog(ComponentDialogUI):
    def __init__(self, parent, component):
        super().__init__(parent)
        self.sizerDlgButtonsOK.SetDefault()
        self.layer = component.layer
        self.baseGlyph = component.baseGlyph
        self.offset = component.offset
        self.scale = component.scale

    @property
    def baseGlyph(self):
        return self.textCtrl_name.Value

    @baseGlyph.setter
    def baseGlyph(self, value):
        self.textCtrl_name.Value = value

    @property
    def offset(self):
        x = self.spinCtrl_pos_x.Value
        y = self.spinCtrl_pos_y.Value
        return x, y

    @offset.setter
    def offset(self, value):
        x, y = value
        self.spinCtrl_pos_x.Value = round(x)
        self.spinCtrl_pos_y.Value = round(y)
        # 		print 'set offset', x, y

    @property
    def scale(self):
        x = self.spinCtrl_scale_x.Value
        y = self.spinCtrl_scale_y.Value
        return x / 100.0, y / 100.0

    @scale.setter
    def scale(self, value):
        x, y = value
        self.spinCtrl_scale_x.Value = round(x * 100.0)
        self.spinCtrl_scale_y.Value = round(y * 100.0)

    def on_button_name(self, event):
        with FindGlyphDialog(self, self.layer) as findGlyphDialog:
            # findGlyphDialog.layer = self.layer
            if (
                findGlyphDialog.ShowModal() == wx.ID_OK
                and findGlyphDialog.selctedGlyph is not None
            ):
                self.baseGlyph = findGlyphDialog.selctedGlyph




# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version 3.9.0 Oct 29 2020)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

from .kerningViewCanvas import KerningViewCanvas
import wx
import wx.xrc

###########################################################################
## Class KerningViewPanelUI
###########################################################################

class KerningViewPanelUI ( wx.Panel ):

    def __init__( self, parent, id = wx.ID_ANY, pos = wx.DefaultPosition, size = wx.Size( 696,300 ), style = wx.TAB_TRAVERSAL, name = u"GlyphViewPanel" ):
        wx.Panel.__init__ ( self, parent, id = id, pos = pos, size = size, style = style, name = name )

        sizer_main = wx.BoxSizer( wx.VERTICAL )

        sizer_controls = wx.BoxSizer( wx.HORIZONTAL )

        self.lbl_font_size = wx.StaticText( self, wx.ID_ANY, u"Font Size", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_font_size.Wrap( -1 )

        sizer_controls.Add( self.lbl_font_size, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.spinCtrl_font_size = wx.SpinCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 0, 1000, 100 )
        sizer_controls.Add( self.spinCtrl_font_size, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.lbl_line_space = wx.StaticText( self, wx.ID_ANY, u"Line Space", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_line_space.Wrap( -1 )

        sizer_controls.Add( self.lbl_line_space, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.spinCtrl_line_space = wx.SpinCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 0, 2000, 120 )
        sizer_controls.Add( self.spinCtrl_line_space, 0, wx.ALL, 5 )

        self.lbl_percent = wx.StaticText( self, wx.ID_ANY, u"%", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_percent.Wrap( -1 )

        sizer_controls.Add( self.lbl_percent, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )


        sizer_controls.Add( ( 0, 0), 1, wx.EXPAND, 5 )

        self.btn_clear = wx.Button( self, wx.ID_ANY, u"Clear", wx.DefaultPosition, wx.DefaultSize, wx.BU_EXACTFIT )
        sizer_controls.Add( self.btn_clear, 0, wx.ALL, 5 )


        sizer_main.Add( sizer_controls, 0, wx.EXPAND, 5 )

        self.kerningViewCanvas = KerningViewCanvas( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.ALWAYS_SHOW_SB|wx.HSCROLL|wx.VSCROLL|wx.WANTS_CHARS, u"KerningViewCanvas" )
        self.kerningViewCanvas.SetScrollRate( 1, 1 )
        sizer_main.Add( self.kerningViewCanvas, 1, wx.EXPAND, 0 )


        self.SetSizer( sizer_main )
        self.Layout()

        # Connect Events
        self.spinCtrl_font_size.Bind( wx.EVT_SPINCTRL, self.on_spinCtrl_font_size )
        self.spinCtrl_line_space.Bind( wx.EVT_SPINCTRL, self.on_spinCtrl_line_space )
        self.btn_clear.Bind( wx.EVT_BUTTON, self.on_btn_clear )

    def __del__( self ):
        pass


    # Virtual event handlers, overide them in your derived class
    def on_spinCtrl_font_size( self, event ):
        event.Skip()

    def on_spinCtrl_line_space( self, event ):
        event.Skip()

    def on_btn_clear( self, event ):
        event.Skip()





"""
dialogItemPicker
===============================================================================
"""
import wx

from .dialogItemPickerUI import DialogItemPickerUI


class DialogItemPicker(DialogItemPickerUI):
    def __init__(
        self,
        parent,
        title="ItemPicker",
        message="Select options",
        options=None,
        selection=None,
        optionsLabel=None,
        selectionLabel=None,
        width = 500,
        height = 400
    ):
        DialogItemPickerUI.__init__(self, parent)
        self.Title = title
        self.label_message.LabelText = message
        if options:
            self.options = options
        if selection:
            self.selection = selection
        if optionsLabel:
            self.itemPicker.label_options.LabelText = optionsLabel
        if selectionLabel:
            self.itemPicker.label_selection.LabelText = selectionLabel

    @property
    def options(self):
        return self.itemPicker.listBox_options.Items

    @options.setter
    def options(self, value):
        self.itemPicker.listBox_options.Items = value

    @property
    def selection(self):
        return self.itemPicker.listBox_selection.Items

    @selection.setter
    def selection(self, value):
        self.itemPicker.listBox_selection.Items = value



"""
layerDialog
===============================================================================

Get name and color for a layer
"""
import wx

from wbDefcon import Color

from .layerDialogUI import LayerDialogUI


class LayerDialog(LayerDialogUI):
    def __init__(self, name=None, color=None):
        super().__init__(wx.GetApp().TopWindow)
        self.textCtrl_name.Hint = "<New Layer>"
        if isinstance(name, str):
            self.textCtrl_name.Value = name
        if isinstance(color, Color):
            self.colourPicker.Colour = color.wx
        self.sizer_buttonOK.SetDefault()
        self.sizer_buttonOK.Bind(wx.EVT_UPDATE_UI, self.onUpdate_ButtonOK)

    @property
    def layer_name(self):
        return self.textCtrl_name.Value.strip()

    @property
    def layer_color(self):
        if self.colourPicker.Colour == wx.WHITE:
            return None
        return Color.from_wx(self.colourPicker.Colour)

    def onUpdate_ButtonOK(self, event):
        event.Enable(bool(self.layer_name))


def getLayerInfo(name=None, color=None):
    with LayerDialog(name, color) as layerDialog:
        if layerDialog.ShowModal() == wx.ID_OK:
            return layerDialog.layer_name, layerDialog.layer_color
    return None, None



"""
misc
===============================================================================

Misc Glyph Commands 
"""
import logging

import wx

from wbDefcon import Color

from .base import GlyphCommand
from .parameter import (
    ParamBoolRequired,
    ParamColour,
    ParamEnumeration,
    ParamStrRequired,
)

log = logging.getLogger(__name__)


class NoteCommand(GlyphCommand):
    name = "Note"
    ADD = 0
    REPLACE = 1
    parameters = [
        ParamStrRequired("t", "Text"),
        ParamEnumeration("m", "Mode", ["Add to note", "Replace note"], ADD),
    ]

    def _execute(self, glyph):
        if glyph.note and self.m == self.ADD:
            glyph.note = glyph.note.strip() + "\n" + self.t
            print("text added")
        else:
            glyph.note = self.t
            print("text set")


class SkipExportCommand(GlyphCommand):
    name = "Skip export"
    parameters = [ParamBoolRequired("s", "Skip", True)]

    def _execute(self, glyph):
        skipExportGlyphs = glyph.font.lib.get("public.skipExportGlyphs", [])
        if self.s:
            if glyph.name in skipExportGlyphs:
                return
            skipExportGlyphs.append(glyph.name)
        else:
            if glyph.name in skipExportGlyphs:
                skipExportGlyphs.remove(glyph.name)
            else:
                return
        glyph.font.lib["public.skipExportGlyphs"] = skipExportGlyphs


class MarkCommand(GlyphCommand):
    name = "Mark glyph"
    parameters = [ParamColour("c", "Colour", wx.WHITE)]

    def _execute(self, glyph):
        if self.c == wx.WHITE:
            glyph.markColor = None
        else:
            glyph.markColor = Color.from_wx(self.c)



"""
generateOTCFFoptionsDialog
===============================================================================
"""
from .generateOTCFFoptionsDialogUI import GenerateOTCFFoptionsDialogUI


class GenerateOTCFFoptionsDialog(GenerateOTCFFoptionsDialogUI):
    @property
    def autohint(self):
        return self.checkBox_autohint.Value

    @property
    def cffcompress(self):
        return self.choice_cffcompress.Selection

    @property
    def removeOverlaps(self):
        return self.checkBox_removeOverlap.Value

    @property
    def useProductionNames(self):
        return self.checkBox_useProductionNames.Value

    @property
    def writeKernFeature(self):
        return self.checkBox_writeKernFeature.Value

    @property
    def writeMarkFeature(self):
        return self.checkBox_writeMarkFeature.Value



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version 3.10.0-4761b0c)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc
import wx.grid

###########################################################################
## Class FontInfoHintingUI
###########################################################################

class FontInfoHintingUI ( wx.ScrolledWindow ):

    def __init__( self, parent, id = wx.ID_ANY, pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.HSCROLL|wx.VSCROLL, name = wx.EmptyString ):
        wx.ScrolledWindow.__init__ ( self, parent, id = id, pos = pos, size = size, style = style, name = name )

        self.SetExtraStyle( wx.WS_EX_VALIDATE_RECURSIVELY )
        self.SetBackgroundColour( wx.SystemSettings.GetColour( wx.SYS_COLOUR_WINDOW ) )

        sizerMain = wx.BoxSizer( wx.VERTICAL )

        boxHead = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"OpenType head Table" ), wx.VERTICAL )

        sizerHead = wx.GridBagSizer( 0, 0 )
        sizerHead.SetFlexibleDirection( wx.HORIZONTAL )
        sizerHead.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

        self.lbl_openTypeHeadLowestRecPPEM = wx.StaticText( self, wx.ID_ANY, u"Lowest Rec PPEM", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_openTypeHeadLowestRecPPEM.Wrap( -1 )

        self.lbl_openTypeHeadLowestRecPPEM.SetMinSize( wx.Size( 170,-1 ) )

        sizerHead.Add( self.lbl_openTypeHeadLowestRecPPEM, wx.GBPosition( 0, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.txtCtrl_openTypeHeadLowestRecPPEM = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadLowestRecPPEM" )
        sizerHead.Add( self.txtCtrl_openTypeHeadLowestRecPPEM, wx.GBPosition( 0, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.lbl_openTypeHeadFlags = wx.StaticText( self, wx.ID_ANY, u"Head Flags", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_openTypeHeadFlags.Wrap( -1 )

        sizerHead.Add( self.lbl_openTypeHeadFlags, wx.GBPosition( 1, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.checkBox_openTypeHeadFlags_0 = wx.CheckBox( self, wx.ID_ANY, u"Baseline for font at y=0", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 0" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_0, wx.GBPosition( 1, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.checkBox_openTypeHeadFlags_1 = wx.CheckBox( self, wx.ID_ANY, u"Left sidebearing point at x=0", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 1" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_1, wx.GBPosition( 2, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_openTypeHeadFlags_2 = wx.CheckBox( self, wx.ID_ANY, u"Instructions may depend on point size", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 2" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_2, wx.GBPosition( 3, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_openTypeHeadFlags_3 = wx.CheckBox( self, wx.ID_ANY, u"Force ppem to integer values for all internal scaler math", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 3" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_3, wx.GBPosition( 4, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_openTypeHeadFlags_4 = wx.CheckBox( self, wx.ID_ANY, u"Instructions may alter advance width", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 4" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_4, wx.GBPosition( 5, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_openTypeHeadFlags_11 = wx.CheckBox( self, wx.ID_ANY, u"Font data is “lossless”", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 11" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_11, wx.GBPosition( 6, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_openTypeHeadFlags_12 = wx.CheckBox( self, wx.ID_ANY, u"Font converted (produce compatible metrics)", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 12" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_12, wx.GBPosition( 7, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_openTypeHeadFlags_13 = wx.CheckBox( self, wx.ID_ANY, u"Font optimized for ClearType", wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeHeadFlags 13" )
        sizerHead.Add( self.checkBox_openTypeHeadFlags_13, wx.GBPosition( 8, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )


        sizerHead.AddGrowableCol( 1 )

        boxHead.Add( sizerHead, 1, wx.EXPAND, 5 )


        sizerMain.Add( boxHead, 0, wx.ALL|wx.EXPAND, 5 )

        boxGasp = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"OpenType gasp Table" ), wx.VERTICAL )

        self.grid_gasp = wx.grid.Grid( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, 0, u"openTypeGaspRangeRecords" )

        # Grid
        self.grid_gasp.CreateGrid( 0, 5 )
        self.grid_gasp.EnableEditing( True )
        self.grid_gasp.EnableGridLines( True )
        self.grid_gasp.EnableDragGridSize( False )
        self.grid_gasp.SetMargins( 0, 0 )

        # Columns
        self.grid_gasp.SetColSize( 0, 100 )
        self.grid_gasp.SetColSize( 1, 100 )
        self.grid_gasp.SetColSize( 2, 100 )
        self.grid_gasp.SetColSize( 3, 100 )
        self.grid_gasp.SetColSize( 4, 100 )
        self.grid_gasp.EnableDragColMove( False )
        self.grid_gasp.EnableDragColSize( True )
        self.grid_gasp.SetColLabelValue( 0, u"Range" )
        self.grid_gasp.SetColLabelValue( 1, u"Gritfit" )
        self.grid_gasp.SetColLabelValue( 2, u"Grayscale" )
        self.grid_gasp.SetColLabelValue( 3, u"Sym. Smoothing" )
        self.grid_gasp.SetColLabelValue( 4, u"Sym. Gritfit" )
        self.grid_gasp.SetColLabelSize( 20 )
        self.grid_gasp.SetColLabelAlignment( wx.ALIGN_CENTER, wx.ALIGN_CENTER )

        # Rows
        self.grid_gasp.EnableDragRowSize( False )
        self.grid_gasp.SetRowLabelSize( 1 )
        self.grid_gasp.SetRowLabelAlignment( wx.ALIGN_CENTER, wx.ALIGN_CENTER )

        # Label Appearance

        # Cell Defaults
        self.grid_gasp.SetDefaultCellAlignment( wx.ALIGN_LEFT, wx.ALIGN_TOP )
        boxGasp.Add( self.grid_gasp, 1, wx.ALL|wx.EXPAND, 5 )


        sizerMain.Add( boxGasp, 1, wx.ALL|wx.EXPAND, 5 )


        self.SetSizer( sizerMain )
        self.Layout()
        sizerMain.Fit( self )

    def __del__( self ):
        pass





"""
toolerase
===============================================================================

Implementation of the erase tool for the glyph editor
"""
from wbDefcon import Anchor, Component, Guideline, Point

from ..cursor import cursors
from .base import IDLE, GlyphTool


class ToolErase(GlyphTool):
    """
    Tool to erase points, anchors, components, guidelines
    """
    cursor = cursors["Erase"]

    def setCursor(self):
        if self.subject is None:
            self.canvas.SetCursor(cursors["Erase"])
        elif isinstance(self.subject, Point):
            self.canvas.SetCursor(cursors["ErasePoint"])
        elif isinstance(self.subject, Anchor):
            self.canvas.SetCursor(cursors["EraseAnchor"])
        elif isinstance(self.subject, Component):
            self.canvas.SetCursor(cursors["EraseComponent"])
        elif isinstance(self.subject, Guideline):
            self.canvas.SetCursor(cursors["EraseGuideline"])

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_LEFT_DOWN(self, event):
        super().on_LEFT_DOWN(event)
        glyph = self.glyph
        if glyph is not None and self.subject is not None:
            glyph.undoManager.saveState()
            if isinstance(self.subject, Point):
                point = self.subject
                for contour in glyph:
                    segmentIndex = contour.segmentIndex(point)
                    if segmentIndex >= 0:
                        if point.segmentType is None:
                            # erase off-curve point, convert to line
                            segment = contour.segments[segmentIndex]
                            segment[-1].segmentType = "line"
                            segment[-1].smooth = False
                            for p in segment[:-1]:
                                contour.removePoint(p)
                        else:
                            # erase on-curve point, remove the segment
                            contour.removeSegment(segmentIndex, True)
                            if len(contour) == 0:
                                glyph.removeContour(contour)
                        self.contoursChanged = True
            elif isinstance(self.subject, Anchor):
                glyph.removeAnchor(self.subject)
                self.anchorsChanged = True
            elif isinstance(self.subject, Component):
                glyph.removeComponent(self.subject)
                self.componentsChanged = True
            elif isinstance(self.subject, Guideline):
                if self.subject.glyph is None:
                    glyph.font.removeGuideline(self.subject)
                    self.fontGuidelinesChanged = True
                else:
                    glyph.removeGuideline(self.subject)
                    self.glyphGuidelinesChanged = True
            if self.contoursChanged or self.componentsChanged:
                glyph.destroyRepresentation("outlineErrors")
            self.postNotifications()
            self.canvas.Refresh()
        self.state = IDLE
        event.Skip()

    def on_MOTION(self, event):
        super().on_MOTION(event)
        if self.glyph is not None:
            if self.state == IDLE:
                self.hitTest()
                self.setCursor()
        event.Skip()



"""
document
===============================================================================
"""
from __future__ import annotations

import logging
import os
import subprocess
import sys
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Optional

import glyphsLib
import wbDefcon
import wx
from fontTools.misc import plistlib
from fontTools.ufoLib import UFOFileStructure, UFOFormatVersion, UFOReader
from fontTools.ufoLib.errors import UFOLibError
from wbBase.document import Document, FileNameFromPath, dbg
from wbDefcon import Font
from wbDefcon.tools.conversion import robofont_guides_to_UFO3, robofont_mark_to_UFO3
from wbFontParts import RFont

# from .dialog.newFontDialog import NewFontDialog
from .dialog.revertFontDialog import RevertFontDialog
from .view.font import UfoFontView
from .view.glyph import UfoGlyphView

if TYPE_CHECKING:
    from wbDefcon import Glyph
    from defcon.tools.notifications import Notification
    from .template import UfoTemplate

log = logging.getLogger(__name__)

if hasattr(subprocess, "STARTUPINFO"):
    info = subprocess.STARTUPINFO()
    info.dwFlags = subprocess.STARTF_USESHOWWINDOW
    info.wShowWindow = subprocess.SW_HIDE
else:
    info = None


class UFOsaveProgressBar(wx.ProgressDialog):
    def __init__(self, font:Font, formatVersion=None, structure=None, path=None):
        tickCount = font.getSaveProgressBarTickCount(formatVersion, structure, path)
        # print("tickCount", tickCount)
        wx.ProgressDialog.__init__(
            self,
            title="Save UFO",
            message="",
            maximum=tickCount,
            parent=wx.GetApp().TopWindow,
            style=wx.PD_APP_MODAL | wx.PD_AUTO_HIDE,
        )

    def __enter__(self):
        return self

    def __exit__(self, type, value, trace_back):
        self.Destroy()

    def update(self, text="", increment=1):
        # print(self.Value + increment, text)
        newVal = min(self.Value + increment, self.Range)
        self.Update(newVal, text)


class UFOsaveProgress:
    def __init__(self, font:Font, formatVersion=None, structure=None, path=None):
        self.tickCount = font.getSaveProgressBarTickCount(
            formatVersion, structure, path
        )
        self.dlg = None

    def __enter__(self):
        self.dlg = wx.ProgressDialog(
            title="Save UFO",
            message="",
            maximum=self.tickCount,
            parent=wx.GetApp().TopWindow,
            style=wx.PD_APP_MODAL | wx.PD_AUTO_HIDE,
        )
        return self.dlg

    def __exit__(self, type, value, trace_back):
        self.dlg.Destroy()
        return True


class UfoDocument(Document):
    """
    Document implementation for Unified Font Objects
    """

    binaryData = False
    canReload = True

    def __init__(self, template:UfoTemplate):
        super().__init__(template)
        self._rFont = None

    @property
    def font(self) -> Font:
        return self._data

    @font.setter
    def font(self, font: Font):
        assert isinstance(font, Font)
        assert self._data is None, "Can not replace font in UfoDoc"
        self._data = font
        self._data.document = self
        # self._rFont = RFont(self._data)

    @font.deleter
    def font(self):
        font = self._data
        if font is None:
            return
        assert isinstance(font, Font)
        del font._doc
        font.endSelfNotificationObservation()
        font.destroyAllRepresentations()
        glyph:Glyph
        for glyph in font:
            glyph.endSelfNotificationObservation()
            glyph.destroyAllRepresentations()
        glyphNames = list(font.keys())
        for name in glyphNames:
            del font[name]
        self._rFont = None
        self._data = None

    @property
    def RFont(self) -> RFont:
        if self._rFont is None and self._data is not None:
            self._rFont = RFont(self._data)
        assert isinstance(self._rFont, RFont)
        return self._rFont

    @property
    def title(self) -> str:
        if self.font is not None:
            if self.font.path:
                name = os.path.basename(self.font.path)
            else:
                name = " ".join(
                    [
                        n
                        for n in (self.font.info.familyName, self.font.info.styleName)
                        if n
                    ]
                )
            return f"Font - {name}"
        else:
            return "No Font"

    @title.setter
    def title(self, title):
        pass
        # raise NotImplementedError

    @property
    def printableName(self) -> str:
        return self.title

    @property
    def modified(self) -> bool:
        """
        True if the document has been saved.
        """
        if self.font is not None:
            return self.font.dirty
        return False

    @modified.setter
    def modified(self, modified):
        _modified = bool(modified)
        if self.font and _modified != self.font.dirty:
            self.font.dirty = _modified
            self.UpdateAllViews(self, ("modify", self, self._modified))

    @property
    def fontView(self) -> Optional[UfoFontView]:
        for view in self.views:
            if isinstance(view, UfoFontView):
                return view
        return None

    # -----------------------------------------------------------------------------
    # internal methods
    # -----------------------------------------------------------------------------

    def beginObservation(self):
        font = self.font
        if font is not None:
            if not font.hasObserver(self, "Font.Changed"):
                font.addObserver(self, "handleNotification", "Font.Changed")
            if not font.info.hasObserver(self, "Info.Changed"):
                font.info.addObserver(self, "handleNotification", "Info.Changed")
            if not font.info.hasObserver(self, "Info.ValueChanged"):
                font.info.addObserver(self, "handleNotification", "Info.ValueChanged")

    def OnNewDocument(self, familyName=None, styleName=None):
        dbg("UfoDocument.OnNewDocument()", indent=1)
        font = Font()
        font.info.unitsPerEm = 1000
        if isinstance(familyName, str):
            font.info.familyName = familyName
        else:
            font.info.familyName = "Untitled"
        if isinstance(styleName, str):
            font.info.styleName = styleName
        font.dirty = False
        self.font = font
        # self.modified = False
        super().OnNewDocument()
        self.beginObservation()
        self.UpdateAllViews(self, ("font loaded", True))
        dbg("UfoDocument.OnNewDocument() - done", indent=0)

    def OnOpenDocument(self, filename: str) -> bool:
        dbg(f'UfoDocument.OnOpenDocument(filename="{filename}")', indent=1)
        result = False
        if self.OnSaveModified():
            if self.typeName == "UFO document":
                from .template import GlyphsTemplate, UfoTemplate, VFBTemplate

                if isinstance(self.template, UfoTemplate):
                    result = self.openUFO(filename)
                elif isinstance(self.template, VFBTemplate):
                    result = self.openVFB(filename)
                elif isinstance(self.template, GlyphsTemplate):
                    result = self.openGlyphs(filename)
                else:
                    wx.LogError("Unknown Template type")
                if result:
                    self.set_modificationDate()
                    self.modified = False
                    self.beginObservation()
                    self.UpdateAllViews(self, ("font loaded", True))
        dbg(f"UfoDocument.OnOpenDocument() -> {result}", indent=0)
        return result

    def openUFO(self, filename: str) -> bool:
        dbg(f'UfoDocument.openUFO(filename="{filename}")', indent=1)
        result = False
        ufoPath = self.getUFOpath(filename)
        if ufoPath:
            try:
                self.font = Font(ufoPath)
                self.font.lib.dirty = False  # why is lib dirty here?
                self.path = ufoPath
                self.saved = True
                result = True
            except UFOLibError:
                result = False
        dbg(f"UfoDocument.openUFO() -> {result}")
        dbg(indent=0)
        return result

    def openVFB(self, filename: str) -> bool:
        dbg(f'UfoDocument.openVFB(filename="{filename}")')
        if os.path.isfile(filename):
            steps = 7
            progress = wx.ProgressDialog(
                title="Import VFB",
                message="",
                maximum=steps,
                parent=self.manager.documentNotebook,
                style=wx.PD_APP_MODAL | wx.PD_AUTO_HIDE,
            )
            with TemporaryDirectory() as tempDir:
                progress.Update(0, "Reading VFB file")
                tempUFOpath = os.path.join(tempDir, "temp.ufo")
                log.debug(tempUFOpath)
                arguments = [
                    "vfb2ufo",
                    "-pp",
                    "-s",
                    "-fo",
                    filename,
                    tempUFOpath,
                ]
                process = subprocess.Popen(
                    arguments,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    startupinfo=info,
                )
                outdata, errordata = process.communicate()
                log.debug("vfb2ufo returncode: %s", process.returncode)
                process.terminate()
                if errordata:
                    log.error(errordata.decode())
                    raise Exception("Conversion vfb2ufo failed")
                if outdata:
                    log.info(outdata.decode().strip())
                if not os.path.isdir(tempUFOpath):
                    raise Exception("Temp UFO folder not found")
                progress.Update(2, "Reading converted data")
                # fix groups.plist
                groupsPath = os.path.join(tempUFOpath, "groups.plist")
                if os.path.isfile(groupsPath):
                    with open(groupsPath, "rb") as groupsFile:
                        groups = plistlib.load(groupsFile)
                    for groupName in list(groups.keys()):
                        if groupName.startswith("public.kern"):
                            del groups[groupName]
                    with open(groupsPath, "wb") as groupsFile:
                        plistlib.dump(groups, groupsFile)
                wbDefconFont = Font(tempUFOpath)
                progress.Update(3, "Processing converted data")
                fontPartFont = RFont(wbDefconFont)
                progress.Update(4, "Loading font")
                font:RFont = fontPartFont.copy()
                progress.Update(5, "Closing input file")
                fontPartFont.close()
                fontPartFont = None
                wbDefconFont = None
            progress.Update(6, "Post processing")
            self.font = font.naked()
            robofont_guides_to_UFO3(self.font)
            robofont_mark_to_UFO3(self.font)
            self.font._ufoFormatVersion = UFOFormatVersion.FORMAT_3_0
            self.font._ufoFileStructure = UFOFileStructure.ZIP
            self._rFont = font
            from .template import UfoTemplate

            self._template = self.manager.FindTemplateByType(UfoTemplate)
            progress.Update(7, "Done")
            progress.Destroy()
            self.saved = False
            self._path = ""
            dbg("UfoDocument.openVFB() done")
            return True
        dbg("UfoDocument.openVFB() failed file not found")
        return False

    def openGlyphs(self, filename: str) -> bool:
        dbg(f'UfoDocument.openGlyphs(filename="{filename}")', indent=1)
        result = False
        ufos = glyphsLib.load_to_ufos(filename, ufo_module=wbDefcon)
        if len(ufos) == 1:
            self.font = ufos[0]
            self.font._ufoFormatVersion = UFOFormatVersion.FORMAT_3_0
            self.font._ufoFileStructure = UFOFileStructure.ZIP
            from .template import UfoTemplate

            self._template = self.manager.FindTemplateByType(UfoTemplate)
            self.path = ""
            self.saved = False
            dbg("UfoDocument.openGlyphs() done")
            return True
        dbg(f"UfoDocument.openGlyphs() -> {result}")
        dbg(indent=0)
        return result

    def getUFOpath(self, filename: str) -> str:
        if filename:
            head, tail = os.path.split(filename)
            __, ext = os.path.splitext(tail)
            if ext.upper() in (".UFO", ".UFOZ"):
                return filename
            elif ext.upper() == ".PLIST" and tail in self.validFileNames:
                __, ext = os.path.splitext(head)
                if ext.upper() == ".UFO":
                    return head
        return ""

    def revert(self) -> None:
        if self.path and os.path.exists(self.path):
            font = self.font
            if not self.modificationDateCorrect:
                reader = UFOReader(self.path, validate=font.ufoLibReadValidate)
                layerNames = reader.getLayerNames()
                font._reader = reader
                for layerName in layerNames:
                    if layerName in font.layers:
                        font.layers[layerName]._glyphSet = reader.getGlyphSet(
                            layerName,
                            validateRead=font.layers.ufoLibReadValidate,
                            validateWrite=font.layers.ufoLibWriteValidate,
                        )
            dlg = RevertFontDialog(self.frame, font)
            if dlg.ShowModal() == wx.ID_OK:
                if dlg.checkBox_font_lib.Value:
                    font.reloadLib()
                    font.lib.dirty = False
                if dlg.checkBox_font_info.Value:
                    font.reloadInfo()
                    font.info.dirty = False
                if dlg.checkBox_groups.Value:
                    font.reloadGroups()
                    font.groups.dirty = False
                if dlg.checkBox_kerning.Value:
                    font.reloadKerning()
                    font.kerning.dirty = False
                if dlg.checkBox_features.Value:
                    font.reloadFeatures()
                    font.features.dirty = False
                if dlg.checkBox_images.Value:
                    font.reloadImages(font.images.fileNames)
                if dlg.checkBox_data.Value:
                    font.reloadData(font.data.fileNames)
                log.debug("layerdata: %r", dlg.layerdata)
                font.reloadLayers(dlg.layerdata)
                self.set_modificationDate()
                self.UpdateAllViews(self, ["font loaded"])
                self.saved = True
            dlg.Destroy()

    def OnSaveDocument(self, filename: str) -> bool:
        """
        Takes filename (which must not be empty),
        and calls SaveObject. If SaveObject returns true, the
        document is set to unmodified; otherwise, an error message box is
        displayed.
        """
        dbg(f"UfoDocument.OnSaveDocument(filename={filename})", indent=1)
        if not filename:
            dbg("UfoDocument.OnSaveDocument() no filename return False", indent=0)
            return False
        saved = False
        try:
            saved = self.SaveObject(filename)
        except:
            # for debugging purposes
            import traceback

            traceback.print_exc()
            wx.MessageBox(
                "Could not save '%s'.  %s"
                % (FileNameFromPath(filename), sys.exc_info()[1]),
                "File Error",
                wx.OK | wx.ICON_EXCLAMATION,
                self.frame,
            )
            dbg("UfoDocument.OnSaveDocument() save failed return False", indent=0)
            return False
        if saved:
            self.path = filename
            self.set_modificationDate()
            self.modified = False
            self.saved = True
            dbg("UfoDocument.OnSaveDocument() -> done return True", indent=0)
            wx.LogStatus(f"UfoDocument saved: {self.printableName}")
            return True
        dbg("UfoDocument.OnSaveDocument() -> not saved return False", indent=0)
        return False

    def SaveObject(self, filename: str) -> bool:
        dbg(f"UfoDocument.SaveObject({filename})", indent=1)
        self.font.holdNotifications()
        with UFOsaveProgressBar(self.font, path=filename) as progress:
            self.font.save(filename, progressBar=progress)
        self.font.releaseHeldNotifications()
        dbg("UfoDocument.SaveObject() -> done", indent=0)
        return True

    def handleNotification(self, notification:Notification):
        log.debug("UfoDocument.handleNotification %r", notification.name)
        view = self.fontView
        if view:
            if notification.name == "Info.ValueChanged":
                if notification.data["attribute"] in ("familyName", "styleName"):
                    view.frame.title = self.title
                if self.font.info.dirty:
                    self.font.dirty = True
            if notification.name == "Font.Changed":
                view.OnUpdate(self, ("modify", notification.data))

    def showGlyph(
        self, glyphname: str, newPage: bool = False, view: Optional[UfoGlyphView] = None
    ) -> bool:
        dbg(
            f"UfoDocument.showGlyph(glyphname={glyphname}, newPage={newPage}, view={view})"
        )
        font = self.font
        if font and glyphname in font:
            glyph = font[glyphname]
            newView = False
            if not newPage and view is None:
                for v in self.views:
                    if isinstance(v, UfoGlyphView):
                        view = v
                        break
            if not view:
                view = UfoGlyphView()
                if view.OnCreate(self, None):
                    self.AddView(view)
                    newView = True
                else:
                    wx.LogError("Can not create view for %r" % glyph)
                    view.Destroy()
                    return False
            view.glyph = glyph
            view.Activate()
            view.frame.CurrentPage.Refresh()
            if newView:
                canvas = view.frame.canvas
                canvas.zoom = canvas.ClientSize.height / (font.info.unitsPerEm * 1.2)
                bounds = glyph.bounds
                if bounds:
                    x0, y0, x1, y1 = bounds
                    x = (x0 + x1) / 2
                    y = (y0 + y1) / 2
                else:
                    x = y = 500
                wx.LogDebug("UfoDoc.showGlyph.centerCanvasOnScreen(%r, %r)" % (x, y))
                canvas.centerCanvasOnScreen(x, y)
            return True
        wx.LogError("Can not find glyph %r in font %r" % (glyphname, font))
        return False

    def DeleteContents(self) -> None:
        font = self.font
        if font is not None:
            if font.hasObserver(self, "Font.Changed"):
                font.removeObserver(self, "Font.Changed")
            if font.info.hasObserver(self, "Info.Changed"):
                font.info.removeObserver(self, "Info.Changed")
            font.close()
        super().DeleteContents()



"""
kerningViewCanvas
===============================================================================
"""
import logging

import wx
from fontTools.misc.transform import Transform
from fontTools.pens.transformPen import TransformPen
from wbDefcon import Glyph
from wbDefcon.pens import graphicsPen

log = logging.getLogger(__name__)

class KerningViewCanvas(wx.ScrolledWindow):
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.ALWAYS_SHOW_SB | wx.HSCROLL | wx.VSCROLL,
    ):
        super().__init__(parent, id=id, pos=pos, size=size, style=style, name="KerningViewCanvas")
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.SetBackgroundColour("WHITE")
        self._glyphs = []
        self.glyphPositions = [(0, self.linespace)]
        self._caretIndex = 0
        # Connect Events
        # window events
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SCROLLWIN, self.on_SCROLLWIN)
        # mouse events
        self.Bind(wx.EVT_MOUSEWHEEL, self.on_MOUSEWHEEL)

    # --------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------

    @property
    def fontsize(self):
        return self.Parent.fontsize

    @fontsize.setter
    def fontsize(self, value):
        self.Parent.fontsize = value

    @property
    def linespace(self):
        return self.Parent.fontsize * self.Parent.linespace / 100

    @property
    def descender(self):
        return self.linespace - self.fontsize

    @property
    def glyphs(self):
        for glyph in self._glyphs:
            yield glyph

    @property
    def glyphCount(self):
        return len(self._glyphs)

    # --------------------------------------------------------------------------
    # Public Methods
    # --------------------------------------------------------------------------

    def insertGlyph(self, index, glyph):
        _glyph = None
        if isinstance(glyph, str) and glyph in ("\n", "\r"):
            _glyph = "\n"
        elif isinstance(glyph, Glyph):
            _glyph = glyph
        elif isinstance(glyph, RGlyph):
            _glyph = glyph.naked()
        if _glyph is not None:
            if isinstance(_glyph, Glyph):
                for notificationName in self.refreshNotifications:
                    if not _glyph.hasObserver(self, notificationName):
                        _glyph.addObserver(self, "handleNotification", notificationName)
            self._glyphs.insert(index, _glyph)
            self.caretIndex = index + 1
            self.SetFocus()

    def appendGlyph(self, glyph):
        self.insertGlyph(self.glyphCount, glyph)

    def insertGlyphAtCaret(self, glyph):
        self.insertGlyph(self.caretIndex, glyph)

    def clearGlyphs(self):
        for glyph in self._glyphs:
            if isinstance(glyph, str):
                continue
            for notificationName in self.refreshNotifications:
                glyph.removeObserver(self, notificationName)
        self._glyphs = []
        self.caretIndex = 0

    def popGlyph(self, index):
        if 0 <= index < self.glyphCount:
            glyph = self._glyphs.pop(index)
            if isinstance(glyph, Glyph) and glyph not in self._glyphs:
                for notificationName in self.refreshNotifications:
                    glyph.removeObserver(self, notificationName)
            if self.caretIndex >= index:
                self.caretIndex -= 1
            self.Refresh()
            self.SetFocus()
            return glyph

    def handleNotification(self, notification):
        log.debug("handleNotification: %r", notification)
        self.Refresh()

    # =================================================================
    # Event Handlers
    # =================================================================

    def on_SCROLLWIN(self, event):
        self.Refresh()
        event.Skip()

    def OnDraw(self, dc):
        dc.SetBackground(wx.TheBrushList.FindOrCreateBrush(self.GetBackgroundColour()))
        dc.Clear()
        gc = wx.GraphicsContext.Create(dc)
        gc.SetBrush(wx.BLACK_BRUSH)
        gc.SetPen(wx.TRANSPARENT_PEN)
        for i, glyph in enumerate(self.glyphs):
            if glyph == "\n":
                continue
            scale = self.fontsize / glyph.font.info.unitsPerEm
            graphicsPen.path = gc.CreatePath()
            graphicsPen.glyphSet = glyph.layer
            transform = (
                Transform().translate(*self.glyphPositions[i]).scale(scale, -scale)
            )
            pen = TransformPen(graphicsPen, transform)
            glyph.draw(pen)
            gc.FillPath(graphicsPen.path, wx.WINDING_RULE)

    def on_MOUSEWHEEL(self, event):
        if event.AltDown():
            value = event.GetWheelRotation()
            fontsize = self.fontsize
            if value > 0:
                fontsize += 5
            else:
                fontsize -= 5
            self.fontsize = fontsize
        event.Skip()



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version Oct 26 2018)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class RenameGlyphDialogUI
###########################################################################

class RenameGlyphDialogUI ( wx.Dialog ):

    def __init__( self, parent ):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"Rename Glyph", pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.CAPTION|wx.RESIZE_BORDER|wx.STAY_ON_TOP )

        self.SetSizeHints( wx.Size( 400,-1 ), wx.DefaultSize )

        sizer = wx.BoxSizer( wx.VERTICAL )

        sbSizerNameCurrent = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"Current Name and Unicode" ), wx.VERTICAL )

        fgSizerNameCurrent = wx.FlexGridSizer( 0, 2, 0, 0 )
        fgSizerNameCurrent.AddGrowableCol( 1 )
        fgSizerNameCurrent.SetFlexibleDirection( wx.HORIZONTAL )
        fgSizerNameCurrent.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

        self.labelNameCurrent = wx.StaticText( self, wx.ID_ANY, u"Name", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.labelNameCurrent.Wrap( -1 )

        fgSizerNameCurrent.Add( self.labelNameCurrent, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.textCtrlNameCurrent = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.textCtrlNameCurrent.Enable( False )

        fgSizerNameCurrent.Add( self.textCtrlNameCurrent, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.labelUniCurrent = wx.StaticText( self, wx.ID_ANY, u"Unicode", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.labelUniCurrent.Wrap( -1 )

        fgSizerNameCurrent.Add( self.labelUniCurrent, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.textCtrlUniCurrent = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.textCtrlUniCurrent.Enable( False )

        fgSizerNameCurrent.Add( self.textCtrlUniCurrent, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )


        sbSizerNameCurrent.Add( fgSizerNameCurrent, 0, wx.EXPAND, 5 )


        sizer.Add( sbSizerNameCurrent, 0, wx.ALL|wx.EXPAND, 5 )

        sbSizerNameNew = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"New Name and Unicode" ), wx.VERTICAL )

        fgSizerNameNew = wx.FlexGridSizer( 0, 3, 0, 0 )
        fgSizerNameNew.AddGrowableCol( 1 )
        fgSizerNameNew.SetFlexibleDirection( wx.HORIZONTAL )
        fgSizerNameNew.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

        self.labelNameNew = wx.StaticText( self, wx.ID_ANY, u"Name", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.labelNameNew.Wrap( -1 )

        fgSizerNameNew.Add( self.labelNameNew, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.textCtrlNameNew = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.textCtrlNameNew.SetMaxLength( 63 )
        fgSizerNameNew.Add( self.textCtrlNameNew, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 5 )

        self.buttonAutoName = wx.BitmapButton( self, wx.ID_ANY, wx.NullBitmap, wx.DefaultPosition, wx.DefaultSize, wx.BU_AUTODRAW|0 )

        self.buttonAutoName.SetBitmap( wx.ArtProvider.GetBitmap( wx.ART_TIP, wx.ART_BUTTON ) )
        fgSizerNameNew.Add( self.buttonAutoName, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.labelUniNew = wx.StaticText( self, wx.ID_ANY, u"Unicode", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.labelUniNew.Wrap( -1 )

        fgSizerNameNew.Add( self.labelUniNew, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.textCtrlUniNew = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        fgSizerNameNew.Add( self.textCtrlUniNew, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 5 )

        self.buttonAutoUni = wx.BitmapButton( self, wx.ID_ANY, wx.NullBitmap, wx.DefaultPosition, wx.DefaultSize, wx.BU_AUTODRAW|0 )

        self.buttonAutoUni.SetBitmap( wx.ArtProvider.GetBitmap( wx.ART_TIP, wx.ART_BUTTON ) )
        fgSizerNameNew.Add( self.buttonAutoUni, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )


        sbSizerNameNew.Add( fgSizerNameNew, 0, wx.EXPAND, 5 )


        sizer.Add( sbSizerNameNew, 0, wx.ALL|wx.EXPAND, 5 )

        sbSizerOptions = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"Options" ), wx.VERTICAL )

        bSizer8 = wx.BoxSizer( wx.VERTICAL )

        self.checkBoxReplaceExisting = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Replace existing glyph with the same name", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxReplaceExisting, 0, wx.ALL|wx.EXPAND, 5 )

        self.checkBoxKeepReplaced = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Keep replaced glyph with the new name", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxKeepReplaced, 0, wx.ALL|wx.EXPAND, 5 )

        radioBoxKeepReplacedChoices = [ u"Keep replaced glyph with new name", u"Remove replaced glyph from composites, kerning, groups and all layers" ]
        self.radioBoxKeepReplaced = wx.RadioBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Replaced glyph", wx.DefaultPosition, wx.DefaultSize, radioBoxKeepReplacedChoices, 1, wx.RA_SPECIFY_COLS )
        self.radioBoxKeepReplaced.SetSelection( 0 )
        bSizer8.Add( self.radioBoxKeepReplaced, 0, wx.ALL, 5 )

        self.staticline = wx.StaticLine( sbSizerOptions.GetStaticBox(), wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        bSizer8.Add( self.staticline, 0, wx.BOTTOM|wx.EXPAND|wx.TOP, 5 )

        self.checkBoxInComposites = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Rename glyph in all composites", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxInComposites, 0, wx.ALL|wx.EXPAND, 5 )

        self.checkBoxInKerning = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Rename glyph in all kerning pairs", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxInKerning, 0, wx.ALL|wx.EXPAND, 5 )

        self.checkBoxInGroups = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Rename glyph in all groups", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxInGroups, 0, wx.ALL|wx.EXPAND, 5 )

        self.checkBoxInFeatures = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Rename glyph in feature code", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxInFeatures, 0, wx.ALL|wx.EXPAND, 5 )

        self.checkBoxAllLayers = wx.CheckBox( sbSizerOptions.GetStaticBox(), wx.ID_ANY, u"Rename glyph on all layers", wx.DefaultPosition, wx.DefaultSize, 0 )
        bSizer8.Add( self.checkBoxAllLayers, 0, wx.ALL, 5 )


        sbSizerOptions.Add( bSizer8, 0, wx.EXPAND, 5 )


        sizer.Add( sbSizerOptions, 0, wx.ALL|wx.EXPAND, 5 )

        buttonSizer = wx.StdDialogButtonSizer()
        self.buttonSizerOK = wx.Button( self, wx.ID_OK )
        buttonSizer.AddButton( self.buttonSizerOK )
        self.buttonSizerCancel = wx.Button( self, wx.ID_CANCEL )
        buttonSizer.AddButton( self.buttonSizerCancel )
        buttonSizer.Realize();

        sizer.Add( buttonSizer, 0, wx.ALL|wx.EXPAND, 5 )


        self.SetSizer( sizer )
        self.Layout()
        sizer.Fit( self )

        self.Centre( wx.BOTH )

        # Connect Events
        self.buttonAutoName.Bind( wx.EVT_BUTTON, self.on_buttonAutoName )
        self.buttonAutoName.Bind( wx.EVT_UPDATE_UI, self.update_buttonAutoName )
        self.buttonAutoUni.Bind( wx.EVT_BUTTON, self.on_buttonAutoUni )
        self.buttonAutoUni.Bind( wx.EVT_UPDATE_UI, self.update_buttonAutoUni )
        self.checkBoxReplaceExisting.Bind( wx.EVT_UPDATE_UI, self.update_checkBoxReplaceExisting )
        self.checkBoxKeepReplaced.Bind( wx.EVT_UPDATE_UI, self.update_checkBoxKeepReplaced )
        self.radioBoxKeepReplaced.Bind( wx.EVT_UPDATE_UI, self.update_radioBoxKeepReplaced )
        self.checkBoxInComposites.Bind( wx.EVT_UPDATE_UI, self.update_checkBoxInComposites )
        self.checkBoxInKerning.Bind( wx.EVT_UPDATE_UI, self.update_checkBoxInKerning )
        self.checkBoxInGroups.Bind( wx.EVT_UPDATE_UI, self.update_checkBoxInGroups )
        self.checkBoxInFeatures.Bind( wx.EVT_UPDATE_UI, self.update_checkBoxInFeatures )
        self.buttonSizerOK.Bind( wx.EVT_BUTTON, self.on_buttonOK )

    def __del__( self ):
        pass


    # Virtual event handlers, overide them in your derived class
    def on_buttonAutoName( self, event ):
        event.Skip()

    def update_buttonAutoName( self, event ):
        event.Skip()

    def on_buttonAutoUni( self, event ):
        event.Skip()

    def update_buttonAutoUni( self, event ):
        event.Skip()

    def update_checkBoxReplaceExisting( self, event ):
        event.Skip()

    def update_checkBoxKeepReplaced( self, event ):
        event.Skip()

    def update_radioBoxKeepReplaced( self, event ):
        event.Skip()

    def update_checkBoxInComposites( self, event ):
        event.Skip()

    def update_checkBoxInKerning( self, event ):
        event.Skip()

    def update_checkBoxInGroups( self, event ):
        event.Skip()

    def update_checkBoxInFeatures( self, event ):
        event.Skip()

    def on_buttonOK( self, event ):
        event.Skip()





"""
addellipse
===============================================================================
"""

from wbDefcon import Contour, Point

from ..cursor import cursors
from .base import IDLE, RUBBERBANDING, AddShapeTool


class AddEllipse(AddShapeTool):
    cursor = cursors["DrawEllipse"]

    def on_LEFT_UP(self, event):
        if self.glyph is not None:
            x, y, w, h = self.dimension
            if w and h:
                left = self.canvas.screenToCanvasX(x)
                top = self.canvas.screenToCanvasY(y)
                right = self.canvas.screenToCanvasX(x + w)
                bottom = self.canvas.screenToCanvasY(y + h)
                bcp_x = self.canvas.screenToCanvasXrel(w) * 0.276
                bcp_y = self.canvas.screenToCanvasYrel(h) * -0.276
                center_x = (left + right) / 2.0
                center_y = (bottom + top) / 2.0
                ellipse = Contour()
                # left to bottom
                ellipse.appendPoint(Point((left, center_y - bcp_y)))
                ellipse.appendPoint(Point((center_x - bcp_x, bottom)))
                ellipse.appendPoint(Point((center_x, bottom), "curve", True))
                # bottom to right
                ellipse.appendPoint(Point((center_x + bcp_x, bottom)))
                ellipse.appendPoint(Point((right, center_y - bcp_y)))
                ellipse.appendPoint(Point((right, center_y), "curve", True))
                # right to top
                ellipse.appendPoint(Point((right, center_y + bcp_y)))
                ellipse.appendPoint(Point((center_x + bcp_x, top)))
                ellipse.appendPoint(Point((center_x, top), "curve", True))
                # top to left
                ellipse.appendPoint(Point((center_x - bcp_x, top)))
                ellipse.appendPoint(Point((left, center_y + bcp_y)))
                ellipse.appendPoint(Point((left, center_y), "curve", True))
                ellipse.selected = True
                ellipse.round()
                self.unselectAll()
                self.glyph.appendContour(ellipse)
        AddShapeTool.on_LEFT_UP(self, event)

    def draw(self, gc):
        if self.state == RUBBERBANDING and not None in (
            self.xClick,
            self.yClick,
            self.x,
            self.y,
        ):
            gc.SetPen(self.pen)
            gc.SetBrush(self.brush)
            gc.DrawEllipse(*self.dimension)



import logging
import os

import wx

from wbDefcon import Font
from wbBase.document import dbg
from wbBase.document.notebook import DocumentPageMixin
from wbBase.document.view import View

from ..font.feature import FeaturePage
from ..font.groups import GroupsPage
from ..font.info import InfoPage
from ..font.kerning import KerningPage

log = logging.getLogger(__name__)


class UfoFontInfoWindow(wx.Notebook, DocumentPageMixin):
    """
    Document Window for UFO Font Info
    """

    def __init__(self, parent, doc, view):
        ID = wx.ID_ANY
        pos = wx.DefaultPosition
        size = wx.DefaultSize
        wx.Notebook.__init__(
            self,
            parent,
            ID,
            pos,
            size,
            style=wx.NB_FIXEDWIDTH | wx.NO_BORDER,
            name="UfoFontWindow",
        )
        DocumentPageMixin.__init__(self, doc, view)
        self.SetExtraStyle(wx.WS_EX_VALIDATE_RECURSIVELY)
        self._font = None
        notebookIndex = 0
        tabIcons = wx.ImageList(16, 16)
        self.AssignImageList(tabIcons)

        # # GlyphGrid page
        # self.glyphGridPage = GlyphGridPage(self)
        # self.glyphGridPanel = self.glyphGridPage.glyphGridPanel
        # self.AddPage(self.glyphGridPage, "Glyphs", True)
        # icon = wx.ArtProvider.GetBitmap("VIEW_ICON", wx.ART_FRAME_ICON)
        # if icon.IsOk():
        #     tabIcons.Add(icon)
        #     self.SetPageImage(notebookIndex, notebookIndex)
        #     notebookIndex += 1

        # Info page
        self.infoPage = InfoPage(self)
        self.AddPage(self.infoPage, "Info", False)
        icon = wx.ArtProvider.GetBitmap("VIEW_INFO", wx.ART_FRAME_ICON)
        if icon.IsOk():
            tabIcons.Add(icon)
            self.SetPageImage(notebookIndex, notebookIndex)
            notebookIndex += 1

        # Feature page
        self.featurePage = FeaturePage(self)
        self.AddPage(self.featurePage, "Feature", False)
        icon = wx.ArtProvider.GetBitmap("BLANK", wx.ART_FRAME_ICON)
        if icon.IsOk():
            tabIcons.Add(icon)
            self.SetPageImage(notebookIndex, notebookIndex)
            notebookIndex += 1

        # Groups page
        self.groupsPage = GroupsPage(self)
        self.AddPage(self.groupsPage, "Groups", False)
        icon = wx.ArtProvider.GetBitmap("VIEW_GROUP", wx.ART_FRAME_ICON)
        if icon.IsOk():
            tabIcons.Add(icon)
            self.SetPageImage(notebookIndex, notebookIndex)
            notebookIndex += 1

        # Kerning page
        self.kerningPage = KerningPage(self)
        self.AddPage(self.kerningPage, "Kerning", False)
        icon = wx.ArtProvider.GetBitmap("VIEW_KERNING", wx.ART_FRAME_ICON)
        if icon.IsOk():
            tabIcons.Add(icon)
            self.SetPageImage(notebookIndex, notebookIndex)
            notebookIndex += 1

        # # Connect Events
        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGING, self.on_PAGE_CHANGING)
        #  Window
        self.Bind(wx.EVT_SET_FOCUS, self.on_SET_FOCUS)

    def __repr__(self):
        return f"<{self.__class__.__name__} of {self._font}>"

    # -------------------------------------------------------------------------
    # properties
    # -------------------------------------------------------------------------

    @property
    def font(self):
        """The wbDefcon Font object of this UfoFontWindow"""
        return self._font

    @font.setter
    def font(self, value):
        if value == self._font:
            return
        assert self._font is None
        assert isinstance(
            value, Font
        ), f"Expected Font, got {value}, type {type(value)}"
        assert value == self._document.font
        self._font = value
        self.title = self._document.title.replace("Font", "Info", 1)
        log.debug("Font %r set for %r", self._font, self)
        for page in self.Children:
            page.font = value

    @font.deleter
    def font(self):
        for page in self.Children:
            del page.font

    def Destroy(self):
        del self.font
        return super().Destroy()

    # -------------------------------------------------------------------------
    # event handler
    # -------------------------------------------------------------------------

    def on_PAGE_CHANGING(self, event):
        page = self.GetPage(event.OldSelection)
        if page.Validate():
            if not page.TransferDataFromWindow():
                log.debug("page %r TransferDataFromWindow failed", page)
            event.Skip()
        else:
            log.debug("page %r Validate failed", page)
            event.Veto()

    def on_SET_FOCUS(self, event):
        self.CurrentPage.SetFocus()
        event.Skip()


class UfoFontInfoView(View):
    frameType = UfoFontInfoWindow

    def __init__(self):
        super().__init__()
        self._typeName = "UFO Font Info View"

    def __repr__(self):
        if self._document:
            return f'<{self.__class__.__name__} of "{self._document.printableName}">'
        return f"<{self.__class__.__name__} of {self._document.__class__.__name__}>"

    @property
    def font(self):
        return self.document.font

    def OnCreate(self, doc, flags):
        dbg(f'UfoFontView.OnCreate(doc="{doc}", flags={flags})', indent=1)
        self._document = doc
        # self._typeName = doc.template.viewTypeName
        self._frame = self.frameType(self.documentNotebook, doc, self)
        self.documentNotebook.AddPage(
            self._frame,
            doc.printableName.replace("Font", "Info", 1),
            True,
            wx.ArtProvider.GetBitmap("I", wx.ART_FRAME_ICON),
        )
        dbg(indent=0)
        return True

    def OnUpdate(self, sender, hint):
        dbg(f"UfoFontView.OnUpdate(sender={sender}, hint={hint})", indent=1)
        super().OnUpdate(sender, hint)
        if hint and hint[0] == "font loaded":
            frame = self._frame
            if frame.font is None:
                frame.font = self.document.font
            frame.title = self._document.title.replace("Font", "Info", 1)
            if frame and hasattr(frame, "OnTitleIsModified"):
                frame.OnTitleIsModified()
            frame.TransferDataToWindow()
            # frame.glyphGridPanel.SendSizeEvent()
        dbg(indent=0)

    # def OnClose(self, deleteWindow=True):
    #     dbg(f"UfoFontView.OnClose(deleteWindow={deleteWindow})", indent=1)
    #     result = False
    #     for view in self._document.views:
    #         if view is not self:
    #             dbg(f"UfoFontView.OnClose() closing child view {view}")
    #             if not view.Close(deleteWindow=True):
    #                 dbg(
    #                     f"UfoFontView.OnClose() returns {result}, can not close {view}",
    #                     indent=0,
    #                 )
    #                 return result
    #             dbg("UfoFontView.OnClose() closing child view done")
    #     result = super().OnClose(deleteWindow=deleteWindow)
    #     dbg(indent=0)
    #     return result



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version Oct 26 2018)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class NewGlyphDialogUI
###########################################################################

class NewGlyphDialogUI ( wx.Dialog ):

	def __init__( self, parent ):
		wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"New Glyph", pos = wx.DefaultPosition, size = wx.DefaultSize, style = wx.CAPTION|wx.RESIZE_BORDER )

		self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )

		sizer = wx.BoxSizer( wx.VERTICAL )

		sizerForm = wx.FlexGridSizer( 0, 2, 0, 0 )
		sizerForm.AddGrowableCol( 1 )
		sizerForm.SetFlexibleDirection( wx.BOTH )
		sizerForm.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.label_name = wx.StaticText( self, wx.ID_ANY, u"Name", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label_name.Wrap( -1 )

		sizerForm.Add( self.label_name, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.textCtrl_name = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
		sizerForm.Add( self.textCtrl_name, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 5 )

		self.label_unicode = wx.StaticText( self, wx.ID_ANY, u"Unicode", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label_unicode.Wrap( -1 )

		sizerForm.Add( self.label_unicode, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.textCtrl_unicode = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
		sizerForm.Add( self.textCtrl_unicode, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 5 )


		sizer.Add( sizerForm, 1, wx.EXPAND, 5 )

		sizerButtons = wx.StdDialogButtonSizer()
		self.sizerButtonsOK = wx.Button( self, wx.ID_OK )
		sizerButtons.AddButton( self.sizerButtonsOK )
		self.sizerButtonsCancel = wx.Button( self, wx.ID_CANCEL )
		sizerButtons.AddButton( self.sizerButtonsCancel )
		sizerButtons.Realize();

		sizer.Add( sizerButtons, 0, wx.ALL|wx.EXPAND, 5 )


		self.SetSizer( sizer )
		self.Layout()
		sizer.Fit( self )

		self.Centre( wx.BOTH )

	def __del__( self ):
		pass





import logging

import fontTools.feaLib.ast as ast
import wx
import wx.stc as stc
from feaASTools.inspect import isFeatureAutomatic, setFeatureAutomatic
from wbDefcon import Font
from wbBase.scripting import MacroButtonMixin
from wbBase.control.textEditControl import STCfindReplaceMixin

log = logging.getLogger(__name__)

feaKeywords = """
anchor anchorDef anon anonymous by contour cursive device 
enum enumerate excludeDFLT exclude_dflt feature featureNames from 
ignore IgnoreBaseGlyphs IgnoreLigatures IgnoreMarks 
MarkAttachmentType UseMarkFilteringSet include includeDFLT 
include_dflt language languagesystem lookup lookupflag 
markClass nameid NULL parameters pos position required 
RightToLeft rsub reversesub script sub substitute subtable 
table useExtension valueRecordDef
""".split()

# style IDs for Feature Lexer
STC_FEA_DEFAULT = 0
STC_FEA_COMMENTLINE = 1
STC_FEA_NUMBER = 2
STC_FEA_WORD = 5
STC_FEA_CLASSNAME = 8


class FeatureTextLexer:
    """Lexer for the FeatureTextCtrl"""

    def styleText(self, event):
        buffer = event.EventObject
        lastStyled = buffer.GetEndStyled()
        startPos = max(lastStyled, 0)
        endPos = event.GetPosition()
        curWord = ""
        comment = False
        while startPos < endPos:
            c = chr(buffer.GetCharAt(startPos))
            curWord += c
            if c == "#":
                comment = True
            elif c in ("\n", "\r"):
                comment = False

            if comment:
                # buffer.StartStyling(startPos, 0x1F)
                buffer.StartStyling(startPos)
                buffer.SetStyling(1, STC_FEA_COMMENTLINE)
                curWord = ""
            elif c.isspace() or c in "()[]{};'":
                curWord = curWord.strip()
                if curWord.startswith("@"):
                    style = STC_FEA_CLASSNAME
                elif curWord in feaKeywords:
                    style = STC_FEA_WORD
                else:
                    style = STC_FEA_DEFAULT

                wordStart = max(0, startPos - len(curWord))
                # buffer.StartStyling(wordStart, 0x1F)
                buffer.StartStyling(wordStart)
                buffer.SetStyling(len(curWord), style)
                buffer.SetStyling(1, STC_FEA_DEFAULT)
                curWord = ""
            startPos += 1
            # else:
            # 	startPos += 1


class FeatureTextCtrl(stc.StyledTextCtrl, STCfindReplaceMixin):
    """
    StyledTextCtrl to display the Feature Code.
    """
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE,
        name="FeatureTextCtrl",
    ):
        super().__init__(parent, id, pos, size, style, name)
        STCfindReplaceMixin.__init__(self, "Find in Features", "Replace in Features")
        self._font = None
        self._lexer = FeatureTextLexer()
        self.SetUseTabs(False)
        self.SetTabWidth(4)
        self.SetIndent(4)
        self.SetTabIndents(True)
        self.SetBackSpaceUnIndents(True)
        self.SetViewEOL(False)
        self.SetViewWhiteSpace(False)
        self.SetMarginWidth(2, 0)
        self.SetIndentationGuides(True)
        self.SetReadOnly(False)
        self.SetMarginType(1, stc.STC_MARGIN_SYMBOL)
        self.SetMarginMask(1, stc.STC_MASK_FOLDERS)
        self.SetMarginWidth(1, 16)
        self.SetMarginSensitive(1, True)
        self.SetProperty("fold", "1")
        self.SetFoldFlags(
            stc.STC_FOLDFLAG_LINEBEFORE_CONTRACTED
            | stc.STC_FOLDFLAG_LINEAFTER_CONTRACTED
        )
        self.SetMarginType(0, stc.STC_MARGIN_NUMBER)
        self.SetMarginWidth(0, self.TextWidth(stc.STC_STYLE_LINENUMBER, "_99999"))
        self.MarkerDefine(stc.STC_MARKNUM_FOLDER, stc.STC_MARK_BOXPLUS)
        self.MarkerSetBackground(stc.STC_MARKNUM_FOLDER, wx.BLACK)
        self.MarkerSetForeground(stc.STC_MARKNUM_FOLDER, wx.WHITE)
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN, stc.STC_MARK_BOXMINUS)
        self.MarkerSetBackground(stc.STC_MARKNUM_FOLDEROPEN, wx.BLACK)
        self.MarkerSetForeground(stc.STC_MARKNUM_FOLDEROPEN, wx.WHITE)
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB, stc.STC_MARK_EMPTY)
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND, stc.STC_MARK_BOXPLUS)
        self.MarkerSetBackground(stc.STC_MARKNUM_FOLDEREND, wx.BLACK)
        self.MarkerSetForeground(stc.STC_MARKNUM_FOLDEREND, wx.WHITE)
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_BOXMINUS)
        self.MarkerSetBackground(stc.STC_MARKNUM_FOLDEROPENMID, wx.BLACK)
        self.MarkerSetForeground(stc.STC_MARKNUM_FOLDEROPENMID, wx.WHITE)
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_EMPTY)
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL, stc.STC_MARK_EMPTY)
        # line wrap
        self.SetWrapMode(stc.STC_WRAP_WORD)
        self.SetWrapIndentMode(stc.STC_WRAPINDENT_INDENT)
        self.SetWrapVisualFlags(stc.STC_WRAPVISUALFLAG_START)
        # selection
        self.SetSelBackground(
            True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        )
        self.SetSelForeground(
            True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        )
        self.SetLexer(stc.STC_LEX_CONTAINER)
        self.StyleSetSpec(
            STC_FEA_DEFAULT, "fore:black,back:white,face:Consolas,size:11"
        )
        self.StyleSetSpec(STC_FEA_WORD, "fore:#0000FF,face:Consolas,size:11")
        self.StyleSetSpec(STC_FEA_CLASSNAME, "fore:#009900,face:Consolas,size:11")
        self.StyleSetSpec(
            STC_FEA_COMMENTLINE, "fore:#808080,back:#E0FCF0,face:Consolas,size:11"
        )

        self.updateFont = False

        self.Bind(stc.EVT_STC_CHANGE, self.on_CHANGE)
        self.Bind(stc.EVT_STC_STYLENEEDED, self.on_STYLENEEDED)

    @property
    def font(self):
        """The wbDefcon Font object of this FeatureTextCtrl"""
        return self._font

    @font.setter
    def font(self, value):
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.GrandParent._font
        self._font = value
        self.updateFont = False
        if value.features and value.features.text:
            self.SetText(value.features.text)
        else:
            self.SetText("")
        self.SetModified(False)
        value.features.addObserver(self, "handleNotification", "Features.TextChanged")
        log.debug("FeatureTextCtrl.@font.setter: %s", self._font)
        self.updateFont = True

    @font.deleter
    def font(self):
        if self._font is not None:
            self._font.features.removeObserver(self, "Features.TextChanged")
            self._font = None
        log.debug("FeatureTextCtrl.@font.deleter")

    @property
    def features(self):
        if self._font:
            return self._font.features

    def handleNotification(self, notification):
        if notification.name == "Features.TextChanged":
            newText = notification.data["newValue"]
            if self.GetText() != newText:
                self.SetText(newText)

    def on_CHANGE(self, event):
        if self.updateFont:
            wx.LogDebug("FeatureTextCtrl.on_CHANGE")
            text = self.GetText()
            features = self.features
            if text != features.text:
                features.text = text

    def on_STYLENEEDED(self, event):
        self._lexer.styleText(event)


class FeatureCtxMenue(wx.Menu):
    """Context Menu for feature items in the tree control"""

    def __init__(self, feature, treeCtrl):
        super().__init__()
        self.feature = feature
        self.treeCtrl = treeCtrl
        self.font = treeCtrl.font
        automatic = isFeatureAutomatic(feature)
        self.mnu_auto = wx.MenuItem(
            self, wx.ID_ANY, "Automatic", "Automatc generated feature", wx.ITEM_CHECK
        )
        self.Append(self.mnu_auto)
        self.mnu_auto.Check(automatic)
        self.mnu_delete = wx.MenuItem(
            self, wx.ID_ANY, "Delete Feature", wx.EmptyString, wx.ITEM_NORMAL
        )
        self.Append(self.mnu_delete)
        self.mnu_insert = wx.MenuItem(
            self, wx.ID_ANY, "Insert Feature", wx.EmptyString, wx.ITEM_NORMAL
        )
        self.Append(self.mnu_insert)
        self.mnu_update = wx.MenuItem(
            self, wx.ID_ANY, "Update Feature", "Regenerate feature code", wx.ITEM_NORMAL
        )
        self.Append(self.mnu_update)
        self.mnu_update.Enable(automatic)
        # Connect Events
        self.Bind(wx.EVT_MENU, self.on_menuSelection, id=self.mnu_auto.Id)
        self.Bind(wx.EVT_MENU, self.on_menuSelection, id=self.mnu_delete.Id)
        self.Bind(wx.EVT_MENU, self.on_menuSelection, id=self.mnu_insert.Id)
        self.Bind(wx.EVT_MENU, self.on_menuSelection, id=self.mnu_update.Id)

    def on_menuSelection(self, event):
        print(dir(event))
        print(event.EventObject)
        if event.Id == self.mnu_auto.Id:
            print("mnu_auto")
            print(event.IsChecked())
            setFeatureAutomatic(self.feature, event.IsChecked())
            self.font.features.setSyntaxtree(
                self.font.features.getSyntaxtree(useGlyphNames=False)
            )
            self.treeCtrl.buildTree()
        elif event.Id == self.mnu_update.Id:
            if not isFeatureAutomatic(self.feature):
                answer = wx.MessageBox(
                    f"Feature '{self.feature.name}' is not automatc generated!\n\nUpdate aynway?",
                    "Update Feature",
                    wx.YES_NO,
                    wx.GetApp().TopWindow,
                )
                if answer == wx.NO:
                    return
        event.Skip()


class FeatureTreeCtrl(wx.TreeCtrl):
    """
    Tree Control in the Feature Panel
    """
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE
        | wx.TR_HIDE_ROOT
        | wx.TR_LINES_AT_ROOT
        | wx.TR_HAS_BUTTONS
        | wx.TR_TWIST_BUTTONS
        | wx.TR_NO_LINES
        | wx.TR_FULL_ROW_HIGHLIGHT,
    ):
        super().__init__(parent, id, pos, size, style, name="FeatureTreeCtrl")
        size = 16
        imagelist = wx.ImageList(size, size)
        self.imgLangSys = imagelist.Add(
            wx.Bitmap.FromRGBA(
                size, size, red=0x02, green=0x4F, blue=0x73, alpha=wx.ALPHA_OPAQUE
            )
        )
        self.imgFeature = imagelist.Add(
            wx.Bitmap.FromRGBA(
                size, size, red=0xAC, green=0x02, blue=0x13, alpha=wx.ALPHA_OPAQUE
            )
        )
        self.imgLookup = imagelist.Add(
            wx.Bitmap.FromRGBA(
                size, size, red=0xAC, green=0xA4, blue=0x02, alpha=wx.ALPHA_OPAQUE
            )
        )
        self.imgClass = imagelist.Add(
            wx.Bitmap.FromRGBA(
                size, size, red=0x03, green=0x72, blue=0x30, alpha=wx.ALPHA_OPAQUE
            )
        )
        self.SetImageList(imagelist)

        self.imagelist = imagelist
        self._font = None
        self.root = None

        # Connect Events
        self.Bind(wx.EVT_TREE_SEL_CHANGED, self.on_selectionChanged)
        self.Bind(wx.EVT_RIGHT_DOWN, self.on_rightDown)
        self.Bind(wx.EVT_RIGHT_UP, self.on_rightUp)

    @property
    def font(self):
        """The wbDefcon Font object of this FeatureTextCtrl"""
        return self._font

    @font.setter
    def font(self, value):
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.GrandParent._font
        self._font = value
        # todo: start observation
        self.buildTree()

    @font.deleter
    def font(self):
        # todo: stop observation
        self._font = None

    @property
    def syntaxtree(self):
        if self._font:
            return self._font.features.getSyntaxtree(useGlyphNames=False)

    @property
    def textCtrl(self):
        return self.GrandParent.featureTextCtrl

    def appendStatement(self, parent, statement):
        if isinstance(statement, ast.LanguageSystemStatement):
            treeItemId = self.AppendItem(
                parent,
                f"{statement.script}-{statement.language}",
                self.imgLangSys,
                data=statement,
            )
        elif isinstance(statement, ast.FeatureBlock):
            treeItemId = self.AppendItem(
                parent, statement.name, self.imgFeature, data=statement
            )
            for s in statement.statements:
                self.appendStatement(treeItemId, s)
        elif isinstance(statement, ast.LookupBlock):
            treeItemId = self.AppendItem(
                parent, statement.name, self.imgLookup, data=statement
            )
            for s in statement.statements:
                self.appendStatement(treeItemId, s)
        elif isinstance(statement, ast.GlyphClassDefinition):
            treeItemId = self.AppendItem(
                parent, statement.name, self.imgClass, data=statement
            )

    def buildTree(self):
        self.Freeze()
        log.debug("FeatureTreeCtrl.buildTree")
        self.DeleteAllItems()
        tree = self.syntaxtree
        if tree:
            self.root = self.AddRoot("features")
            for statement in tree.statements:
                self.appendStatement(self.root, statement)
            # self.ExpandAllChildren(self.root)
        else:
            log.debug("No syntaxtree")
        self.Thaw()

    # --------------------------------------------------------------------------
    # Event handler
    # --------------------------------------------------------------------------

    def on_selectionChanged(self, event):
        try:
            data = self.GetItemData(event.Item)
        except RuntimeError:
            return
        if data:
            self.textCtrl.GotoLine(data.location[1])
            self.textCtrl.LineUpExtend()
        event.Skip()

    def on_rightDown(self, event):
        position = event.GetPosition()
        item, __ = self.HitTest(position)
        if item:
            self.SelectItem(item)

    def on_rightUp(self, event):
        position = event.GetPosition()
        item, __ = self.HitTest(position)
        if item:
            data = self.GetItemData(item)
            lineNo = data.location[1]
            if isinstance(data, ast.FeatureBlock):
                menu = FeatureCtxMenue(data, self)
                self.PopupMenu(menu)
                menu.Destroy()
            wx.CallAfter(self.textCtrl.ShowLines, lineNo - 5, lineNo + 5)
        # item, __ = self.HitTest(position)
        # if item:
        #     wx.CallAfter(self.SelectItem, item)


class FeatureStatusPanel(wx.Panel, MacroButtonMixin):
    """
    Tree Status Panel in the Feature Panel
    """
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE,
    ):
        super().__init__(parent, id, pos, size, style, name="FeatureStatusPanel")
        sizer_outer = wx.BoxSizer(wx.VERTICAL)
        self.staticline = wx.StaticLine(
            self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL
        )
        sizer_outer.Add(self.staticline, 0, wx.EXPAND, 0)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.buttonSplit = wx.BitmapButton(
            self,
            wx.ID_ANY,
            wx.ArtProvider.GetBitmap(wx.ART_GOTO_LAST, wx.ART_BUTTON),
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.BU_AUTODRAW | 0,
        )
        sizer.Add(self.buttonSplit, 0, wx.ALL, 5)
        sizer.Add((0, 0), 1, wx.EXPAND, 5)
        self.button_macro = wx.Button(
            self, wx.ID_ANY, "▼", wx.DefaultPosition, wx.DefaultSize, wx.BU_EXACTFIT
        )
        self.button_macro.SetBitmap(wx.ArtProvider.GetBitmap("PYTHON", wx.ART_BUTTON))
        sizer.Add(self.button_macro, 0, wx.ALL, 5)
        sizer_outer.Add(sizer, 1, wx.EXPAND, 5)
        self.SetSizer(sizer_outer)
        self.Layout()

        MacroButtonMixin.__init__(
            self, self.button_macro, "_features", self.GrandParent.view
        )
        # Connect Events
        self.buttonSplit.Bind(wx.EVT_BUTTON, self.on_buttonSplit)
        self.buttonSplit.Bind(wx.EVT_UPDATE_UI, self.update_buttonSplit)

    @property
    def featurePanel(self):
        return self.Parent.featurePanel

    def on_buttonSplit(self, event):
        if self.featurePanel.IsSplit():
            self.featurePanel.Unsplit(self.Parent.featureTreeCtrl)
            self.buttonSplit.SetBitmap(
                wx.ArtProvider.GetBitmap("GO_LAST", wx.ART_BUTTON)
            )
        else:
            self.featurePanel.SplitVertically(
                self.Parent.featureTreeCtrl, self.Parent.featureTextCtrl, 200
            )
            self.buttonSplit.SetBitmap(
                wx.ArtProvider.GetBitmap("GO_FIRST", wx.ART_BUTTON)
            )

    def update_buttonSplit(self, event):
        event.Skip()


class FeaturePage(wx.Panel):
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE | wx.TAB_TRAVERSAL,
    ):
        super().__init__(parent, id, pos, size, style, name="FeaturePage")
        self._font = None
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.featurePanel = wx.SplitterWindow(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SP_3D | wx.SP_NOBORDER | wx.BORDER_NONE,
        )
        self.featureTreeCtrl = FeatureTreeCtrl(self.featurePanel)
        self.featureTextCtrl = FeatureTextCtrl(self.featurePanel)

        sizer.Add(self.featurePanel, 1, wx.EXPAND, 0)
        self.featurePanel.SetMinimumPaneSize(20)
        self.featurePanel.SashGravity = 0.5
        self.featurePanel.SplitVertically(self.featureTreeCtrl, self.featureTextCtrl, 0)
        self.featurePanel.Unsplit(self.featureTreeCtrl)

        self.featureStatusPanel = FeatureStatusPanel(self)
        sizer.Add(self.featureStatusPanel, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        self.Layout()

        # Connect Events
        self.featurePanel.Bind(wx.EVT_IDLE, self.featurePanelOnIdle)
        self.featurePanel.Bind(wx.EVT_SPLITTER_UNSPLIT, self.on_Unsplit)

    @property
    def font(self):
        """The wbDefcon Font object of this FeaturePage"""
        return self._font

    @font.setter
    def font(self, value):
        assert self._font is None
        assert isinstance(value, Font)
        # assert value == self.Parent._document._data
        assert value == self.Parent.font
        self._font = value
        self.featureTextCtrl.font = value
        self.featureTreeCtrl.font = value

    @font.deleter
    def font(self):
        del self.featureTextCtrl.font
        del self.featureTreeCtrl.font
        self._font = None

    def featurePanelOnIdle(self, event):
        self.featurePanel.SetSashPosition(0)
        self.featurePanel.Unbind(wx.EVT_IDLE)

    def on_Unsplit(self, event):
        self.featureStatusPanel.buttonSplit.SetBitmap(
            wx.ArtProvider.GetBitmap(wx.ART_GOTO_LAST, wx.ART_BUTTON)
        )



"""
glyphLock
===============================================================================
"""
import wx
from wx import aui

from ..view.glyph import UfoGlyphView
from . import BaseToolbar

contour, component, margin, metric, anchor, zone, guide = range(7)
fontLevel = (metric, zone, guide)
glyphLevel = (margin, anchor, contour, component, guide)


class GlyphLockToolbar(BaseToolbar):
    itemType = wx.ITEM_CHECK

    def __init__(self, parent):
        super().__init__(parent, "GlyphLockToolbar")
        self.appendTool(
            "Lock Margins",
            "MARGIN_LOCK",
            "Lock glyph margins in current glyph editor",
            margin,
        )
        self.appendTool(
            "Lock Metric",
            "METRIC_LOCK",
            "Lock vertical font metric in current glyph editor",
            metric,
        )
        self.appendTool(
            "Lock Zones",
            "ZONES_LOCK",
            "Lock alignment zones in current glyph editor",
            zone,
        )
        self.appendTool(
            "Lock Guides",
            "GUIDE_LOCK",
            "Lock guides in current glyph editor",
            guide,
        )
        self.appendTool(
            "Lock Anchors",
            "ANCHOR_LOCK",
            "Lock anchors in current glyph editor",
            anchor,
        )
        self.appendTool(
            "Lock Contours",
            "OUTLINE_LOCK",
            "Lock contours in current glyph editor",
            contour,
        )
        self.appendTool(
            "Lock Components",
            "COMPONENT_LOCK",
            "Lock components in current glyph editor",
            component,
        )

    def on_Tool(self, event):
        tool = self.FindTool(event.Id)
        i = tool.GetUserData()
        planes = []
        if i in fontLevel:
            fontPlanes = self.currentView.frame.canvas.getDrawingPlaneStack("Font")
            if i == metric:
                planes.append(fontPlanes.get("VerticalMetric"))
            elif i == zone:
                planes.append(fontPlanes.get("AlignmentZones"))
            elif i == guide:
                planes.append(fontPlanes.get("FontGuide"))
        if i in glyphLevel:
            activePlanes = self.currentView.frame.canvas.getActiveDrawingPlaneStack()
            if i == margin:
                planes.append(activePlanes.get("GlyphMetric"))
            elif i == guide:
                planes.append(activePlanes.get("GlyphGuide"))
            elif i == anchor:
                planes.append(activePlanes.get("Anchors"))
            elif i == contour:
                planes.append(activePlanes.get("OnCurvePoints"))
                planes.append(activePlanes.get("OffCurvePoints"))
                planes.append(activePlanes.get("GlyphOutline"))
            elif i == component:
                planes.append(activePlanes.get("GlyphComponent"))
        for plane in planes:
            plane.locked = event.IsChecked()

    def on_update_Tool(self, event):
        view = self.currentView
        if view and view.document:
            if isinstance(view, UfoGlyphView):
                canvas = view.frame.canvas
                tool = self.FindTool(event.Id)
                i = tool.GetUserData()
                not_yet_supported = (margin, metric, zone)
                planes = []
                if i in not_yet_supported:
                    event.Check(True)
                    event.Enable(False)
                    return
                event.Enable(True)
                if i in fontLevel:
                    planes = canvas.getDrawingPlaneStack("Font")
                    if i == guide:
                        event.Check(planes.get("FontGuide").locked)
                elif i in glyphLevel:
                    planes = canvas.getActiveDrawingPlaneStack()
                    if i == metric:
                        event.Check(planes.get("VerticalMetric").locked)
                    elif i == zone:
                        event.Check(planes.get("AlignmentZones").locked)
                    elif i == margin:
                        event.Check(planes.get("GlyphMetric").locked)
                    elif i == guide:
                        event.Check(planes.get("GlyphGuide").locked)
                    elif i == anchor:
                        event.Check(planes.get("Anchors").locked)
                    elif i == contour:
                        event.Check(planes.get("GlyphOutline").locked)
                    elif i == component:
                        event.Check(planes.get("GlyphComponent").locked)
                return
        event.Enable(False)



import pickle

import wx
import wx.grid as gridlib
from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin, TextEditMixin

from wbDefcon import Font
from wbBase.scripting import MacroButtonMixin

from .groupsStatusPanelUI import GroupsStatusPanelUI


class GroupsListDropTarget(wx.DropTarget):
    """
    DropTarget for GroupsListCtrl
    """

    def __init__(self, groupsListCtrl):
        super().__init__()
        self.groupsListCtrl = groupsListCtrl
        self.SetDataObject(wx.CustomDataObject("application.UFO-WB.glyphNames"))
        self.SetDefaultAction(wx.DragCopy)
        self.documentManager = wx.GetApp().documentManager
        self._font = None
        self._glyphs = []

    @property
    def font(self):
        return self._font

    @property
    def glyphs(self):
        return self._glyphs

    def OnEnter(self, x, y, defResult):
        if self.GetData():
            glyphNameData = self.DataObject.GetData()
            if glyphNameData:
                glyphNameDict = pickle.loads(glyphNameData)
                font = None
                for doc in self.documentManager.documents:
                    if (
                        doc.typeName == "UFO document"
                        and id(doc.font) == glyphNameDict["font"]
                    ):
                        font = doc.font
                        break
                if not font:
                    return wx.DragError
                if font != self.groupsListCtrl.font:
                    return wx.DragNone
                self._font = font
                self._glyphs = [font[g] for g in glyphNameDict["glyphs"] if g in font]
                self.groupsListCtrl.SetFocus()
                return defResult
        return wx.DragNone

    def OnDragOver(self, x, y, defResult):
        if self.font is None:
            return wx.DragNone
        itemIndex, flags = self.groupsListCtrl.HitTest(wx.Point(x, y))
        if flags & wx.LIST_HITTEST_ONITEM:
            self.groupsListCtrl.Select(itemIndex)
            self.groupsListCtrl.EnsureVisible(itemIndex)
        return defResult

    def OnLeave(self):
        self._font = None
        self._glyphs = []

    def OnData(self, x, y, defResult):
        if self.font is None:
            return wx.DragNone
        itemIndex, flags = self.groupsListCtrl.HitTest(wx.Point(x, y))
        if flags & wx.LIST_HITTEST_ONITEM:
            groupName = self.groupsListCtrl.GetItem(itemIndex).Text
            groups = self.font.groups
            groups[groupName] = groups[groupName] + [g.name for g in self.glyphs]
            self.groupsListCtrl.Select(itemIndex)
            self.groupsListCtrl.EnsureVisible(itemIndex)
            return defResult
        return wx.DragNone


class GroupsListCtrl(wx.ListCtrl, ListCtrlAutoWidthMixin, TextEditMixin):
    """
    ListCtrl of the Groups page
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.LC_REPORT | wx.LC_SINGLE_SEL | wx.LC_VIRTUAL,
        name="GroupsListCtrl",
    ):
        wx.ListCtrl.__init__(self, parent, id, pos, size, style, name=name)
        self._font = None
        # self._groupNames = []
        ListCtrlAutoWidthMixin.__init__(self)
        self.InsertColumn(0, "Group")
        self.SetItemCount(0)
        TextEditMixin.__init__(self)
        self.SetDropTarget(GroupsListDropTarget(self))

    @property
    def font(self):
        """The wbDefcon Font object of this FeatureTextCtrl"""
        return self._font

    @font.setter
    def font(self, value):
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.GrandParent._font
        self._font = value
        # self._groupNames = self._font.groups.sortedGroupNames
        self.SetItemCount(len(self.groupNames))
        if self.ItemCount:
            self.Select(0)

    @font.deleter
    def font(self):
        # self._groupNames = []
        self.SetItemCount(0)
        self._font = None

    @property
    def groupNames(self):
        if self._font is None:
            return []
        return self._font.groups.sortedGroupNames

    def OnGetItemText(self, item, column):
        return self.groupNames[item]

    def SetVirtualData(self, row, col, text):
        # todo: implement this
        wx.LogWarning(
            f"Method 'SetVirtualData' of GroupsListCtrl not yet implementd\n\n New name '{text}' not set"
        )


class GroupsStatusPanel(GroupsStatusPanelUI, MacroButtonMixin):
    """
    Status Panel of the Groups page
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE,
    ):
        super().__init__(parent, id, pos, size, style, name="GroupsStatusPanel")
        MacroButtonMixin.__init__(
            self, self.button_macro, "_groups", self.GrandParent.view
        )

    def on_choiceGroups(self, event):
        wx.LogWarning(
            f"Method 'on_choiceGroups' of GroupsStatusPanel not yet implementd.\n\n Filter not set"
        )

    def update_choiceGroups(self, event):
        event.Skip()

    def on_button_New(self, event):
        newGroup = wx.GetTextFromUser(
            "Enter name for new group", "Add new Group", "", self
        )
        if newGroup:
            # todo: implement this
            wx.LogWarning(
                f"Method 'on_button_New' of GroupsStatusPanel not yet implementd\n\n New group '{newGroup}' not created"
            )

    def on_button_Delete(self, event):
        event.Skip()

    def update_button_Delete(self, event):
        event.Skip()


class GroupContentTable(gridlib.GridTableBase):
    """
    GroupContentTable of the Groups page
    """

    def __init__(self, group, font):
        super().__init__()
        self.group = group or []
        self.font = font
        self.colLabels = ("Name", "Left Margin", "Width", "Right Margin")
        self.dataTypes = (
            gridlib.GRID_VALUE_STRING,
            gridlib.GRID_VALUE_NUMBER,
            gridlib.GRID_VALUE_NUMBER,
            gridlib.GRID_VALUE_NUMBER,
        )

    def GetNumberCols(self):
        return len(self.colLabels)

    def GetNumberRows(self):
        return len(self.group)

    def GetColLabelValue(self, col):
        return self.colLabels[col]

    def GetRowLabelValue(self, row):
        return ""

    def CanGetValueAs(self, row, col, typeName):
        colType = self.dataTypes[col].split(":")[0]
        if typeName == colType:
            return True
        else:
            return False

    def CanSetValueAs(self, row, col, typeName):
        return self.CanGetValueAs(row, col, typeName)

    def GetValue(self, row, col):
        glyphName = self.group[row]
        if col == 0:
            return glyphName
        if glyphName in self.font:
            glyph = self.font[glyphName]
            if col == 1:
                return glyph.leftMargin
            if col == 2:
                return glyph.width
            if col == 3:
                return glyph.rightMargin
        return None

    def SetValue(self, row, col, value):
        glyphName = self.group[row]
        if glyphName in self.font:
            glyph = self.font[glyphName]
            if col == 1 and glyph.leftMargin is not None:
                glyph.leftMargin = int(value)
            if col == 2:
                glyph.width = int(value)
            if col == 3 and glyph.rightMargin is not None:
                glyph.rightMargin = int(value)
            self.View.Refresh()

    def GetAttr(self, row, col, kind):
        attr = gridlib.GridCellAttr()
        # todo: allow editing of lsb, rsb and width
        if col == 0:
            attr.SetReadOnly()
        if row % 2:
            attr.SetBackgroundColour(wx.Colour("WHITESMOKE"))
        else:
            attr.SetBackgroundColour(wx.Colour("WHITE"))
        if col > 0:
            attr.SetAlignment(wx.ALIGN_RIGHT, wx.ALIGN_CENTER)
        return attr

    def IsEmptyCell(self, row, col):
        return self.GetValue(row, col) is None


class GroupContentGrid(gridlib.Grid):
    """
    GroupContentGrid of the Groups page
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.WANTS_CHARS | wx.NO_BORDER | gridlib.Grid.GridSelectRows,
    ):
        super().__init__(parent, id, pos, size, style, name="GroupContentGrid")
        self.SetColLabelSize(20)
        self.SetRowLabelSize(1)
        self.SetDefaultCellFitMode(gridlib.GridFitMode.Ellipsize(wx.ELLIPSIZE_MIDDLE))
        # bind events
        # self.Bind(gridlib.EVT_GRID_SELECT_CELL, self.on_selectCell)
        self.Bind(gridlib.EVT_GRID_CELL_LEFT_DCLICK, self.on_CellLeftDoubleClick)

    # def SetTable(self, table, takeOwnership=False, selmode=gridlib.Grid.GridSelectCells):
    #     super().SetTable(table, takeOwnership, selmode)

    def on_CellLeftDoubleClick(self, event):
        col = event.Col
        if col == 0:
            row = event.Row
            glyphName = self.Table.GetValue(row, col)
            if glyphName in self.GrandParent.font:
                self.GrandParent.font.showGlyph(glyphName)
        event.Skip()


class GroupsPage(wx.Panel):
    """
    The Groups page
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.TAB_TRAVERSAL,
        name="GroupsPage",
    ):
        super().__init__(parent, id, pos, size, style, name)
        self._font = None
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.groupsPanel = wx.SplitterWindow(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SP_3D | wx.SP_NOBORDER | wx.BORDER_NONE,
        )
        self.groupsListCtrl = GroupsListCtrl(self.groupsPanel)
        self.groupsContent = GroupContentGrid(self.groupsPanel)
        sizer.Add(self.groupsPanel, 1, wx.EXPAND, 0)
        self.groupsPanel.SetMinimumPaneSize(20)
        self.groupsPanel.SashGravity = 0.3
        self.groupsPanel.SplitVertically(self.groupsListCtrl, self.groupsContent, 200)
        # self.groupsPanel.Unsplit(self.groupsListCtrl)

        self.groupsStatusPanel = GroupsStatusPanel(self)
        sizer.Add(self.groupsStatusPanel, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        self.Layout()
        # Connect Events
        self.groupsPanel.Bind(wx.EVT_IDLE, self.groupsPanelOnIdle)
        self.groupsListCtrl.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_GroupSelected)

    @property
    def font(self):
        """The wbDefcon Font object of this GroupsPage"""
        return self._font

    @font.setter
    def font(self, value:Font):
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.Parent.font
        self._font = value
        self.groupsListCtrl.font = value
        self._font.groups.addObserver(
            self,
            "handleNotification",
            notification=self._font.groups.changeNotificationName,
        )

    @font.deleter
    def font(self):
        del self.groupsListCtrl.font
        if self._font is not None and self._font.groups.hasObserver(
            self, self._font.groups.changeNotificationName
        ):
            self._font.groups.removeObserver(
                self, self._font.groups.changeNotificationName
            )
        self._font = None

    @property
    def currentGroupName(self):
        if self.groupsListCtrl.SelectedItemCount:
            itemIndex = self.groupsListCtrl.GetNextSelected(-1)
            return self.groupsListCtrl.GetItem(itemIndex).Text

    def handleNotification(self, notification):
        # print(notification)
        if self._font is not None:
            self.groupsListCtrl.SetItemCount(len(self._font.groups))
        if self.currentGroupName:
            self.setGroupsContentTable(self.currentGroupName)
        self.Refresh()

    def setGroupsContentTable(self, groupName):
        oldTable = self.groupsContent.Table
        if oldTable and hasattr(oldTable, "group"):
            for glyphName in oldTable.group:
                if glyphName in self.font:
                    glyph = self.font[glyphName]
                    if glyph.hasObserver(self, glyph.changeNotificationName):
                        glyph.removeObserver(self, glyph.changeNotificationName)
        newTabel = GroupContentTable(self.font.groups[groupName], self.font)
        for glyphName in newTabel.group:
            if glyphName in self.font:
                glyph = self.font[glyphName]
                if not glyph.hasObserver(self, glyph.changeNotificationName):
                    glyph.addObserver(
                        self,
                        "handleNotification",
                        notification=glyph.changeNotificationName,
                    )
        self.groupsContent.SetTable(
            newTabel,
            True,
            self.groupsContent.GridSelectRows,
        )
        self.groupsContent.Refresh()

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def groupsPanelOnIdle(self, event):
        self.groupsPanel.SetSashPosition(200)
        self.groupsPanel.Unbind(wx.EVT_IDLE)

    def on_GroupSelected(self, event):
        self.setGroupsContentTable(self.font.groups.sortedGroupNames[event.Index])
        event.Skip()



import wx


class ZoneListCtrl(wx.ListCtrl):
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.LC_REPORT | wx.LC_SINGLE_SEL | wx.LC_VIRTUAL,
        name="ZoneListCtrl",
    ):
        wx.ListCtrl.__init__(
            self, parent, id=id, pos=pos, size=size, style=style, name=name
        )
        self.EnableAlternateRowColours()
        self.AppendColumn("")
        self.AppendColumn("Bottom", wx.LIST_FORMAT_RIGHT)
        self.AppendColumn("Top", wx.LIST_FORMAT_RIGHT)
        self.AppendColumn("Width", wx.LIST_FORMAT_RIGHT)
        self.SetColumnWidth(0, 0)
        self.SetColumnWidth(1, 90)
        self.SetColumnWidth(2, 90)
        self.SetColumnWidth(3, 90)
        self.SetItemCount(0)

        self.Bind(wx.EVT_SIZE, self.on_Size)

    @property
    def value(self):
        return self.Parent.Value

    def OnGetItemText(self, row, col):
        if col == 0:
            return ""
        if col == 1:
            return str(self.value[row * 2])
        if col == 2:
            return str(self.value[row * 2 + 1])
        if col == 3:
            return str(self.value[row * 2 + 1] - self.value[row * 2])

    def on_Size(self, event):
        self.SetColumnWidth(1, round(self.Size.Width / 3))
        self.SetColumnWidth(2, round(self.Size.Width / 3))
        self.SetColumnWidth(3, round(self.Size.Width / 3))
        event.Skip()


class StemListCtrl(wx.ListCtrl):
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.LC_REPORT | wx.LC_SINGLE_SEL | wx.LC_VIRTUAL,
        name="StemListCtrl",
    ):
        wx.ListCtrl.__init__(
            self, parent, id=id, pos=pos, size=size, style=style, name=name
        )
        self.EnableAlternateRowColours()
        self.AppendColumn("Std Stem")
        self.AppendColumn("Width", wx.LIST_FORMAT_RIGHT)
        self.SetColumnWidth(0, 90)
        self.SetColumnWidth(1, 180)
        self.SetItemCount(0)
        self.EnableCheckBoxes()

        self.Bind(wx.EVT_LIST_ITEM_CHECKED, self.on_listItemChecked)
        # self.Bind(wx.EVT_LIST_ITEM_UNCHECKED, self.on_listItemUnChecked)
        self.Bind(wx.EVT_SIZE, self.on_Size)

    @property
    def value(self):
        return self.Parent.Value

    @value.setter
    def value(self, value):
        self.Parent.Value = value

    def OnGetItemText(self, row, col):
        if col == 0:
            return ""
        if col == 1:
            return str(self.value[row])

    def OnGetItemIsChecked(self, item):
        return item == 0

    def on_listItemChecked(self, event):
        # print(event.Index)
        stems = self.value[:]
        stdStem = stems.pop(event.Index)
        stems.sort()
        stems.insert(0, stdStem)
        self.value = stems
        # event.Skip()

    # def on_listItemUnChecked(self, event):
    #     event.Skip()

    def on_Size(self, event):
        self.SetColumnWidth(0, round(self.Size.Width / 3))
        self.SetColumnWidth(1, round(self.Size.Width / 3 * 2))
        event.Skip()



"""
assignLayerDialog
===============================================================================
"""
import os

import wx
from .assignLayerDialogUI import AssignLayerDialogUI


class AssignLayerDialog(AssignLayerDialogUI):
    def __init__(self, parent):
        super().__init__(parent)
        self.button_sizerOK.SetDefault()
        self._fonts = None
        self.choice_source_font.Set([f[0] for f in self.fonts])
        self.choice_source_font.Selection = 0
        self.choice_source_layer.Set([l.name for l in self.source_font.layers])
        self.choice_source_layer.Selection = 0
        self.choice_target_font.Set([f[0] for f in self.fonts])
        self.choice_target_font.Selection = 0
        self.choice_target_layer.Set(
            [l.name for l in self.target_font.layers if l.name != "public.default"]
        )
        self.choice_target_layer.Append("<New Layer>")
        self.choice_target_layer.Selection = 0

    @property
    def app(self):
        return wx.GetApp()

    @property
    def fonts(self):
        if not self._fonts:
            self._fonts = []
            for doc in self.app.documentManager.documents:
                if doc.typeName == "UFO document":
                    self._fonts.append((self._getFontDisplayName(doc.font), doc.font))
        return self._fonts

    @property
    def source_font(self):
        return self.fonts[self.choice_source_font.Selection][1]

    @property
    def source_layer_name(self):
        return self.choice_source_layer.StringSelection

    @property
    def target_font(self):
        return self.fonts[self.choice_target_font.Selection][1]

    @target_font.setter
    def target_font(self, font):
        for i, f in enumerate(self.fonts):
            if f[1] == font:
                self.choice_target_font.Selection = i
                break
        self.choice_target_layer.Set(
            [l.name for l in self.target_font.layers if l.name != "public.default"]
        )
        self.choice_target_layer.Append("<New Layer>")
        self.choice_target_layer.Selection = 0

    @property
    def target_layer_name(self):
        return self.choice_target_layer.StringSelection

    @property
    def create_new_glyphs(self):
        return self.checkBox_new_glyphs.Value

    @staticmethod
    def _getFontDisplayName(font):
        name = f"{font.info.familyName} {font.info.styleName}"
        if font.path:
            path = os.path.basename(font.path)
        else:
            path = "Not saved yet"
        return f"{name} | {path}"

    # =========================================================================
    # Event Handler
    # =========================================================================

    def on_choice_source_font(self, event):
        self.choice_source_layer.Set([l.name for l in self.source_font.layers])
        self.choice_source_layer.Selection = 0
        event.Skip()

    def on_choice_source_layer(self, event):
        event.Skip()

    def on_choice_target_font(self, event):
        self.choice_target_layer.Set(
            [l.name for l in self.target_font.layers if l.name != "public.default"]
        )
        self.choice_target_layer.Append("<New Layer>")
        self.choice_target_layer.Selection = 0
        event.Skip()

    def on_choice_target_layer(self, event):
        event.Skip()



"""
glyphShow
===============================================================================
"""
import wx
from wx import aui

from ..view.glyph import UfoGlyphView
from . import BaseToolbar

(
    margin,
    metric,
    anchor,
    zone,
    guide,
    fill,
    oncurve,
    offcurve,
    coordinates,
    redarraow,
    layer,
) = range(11)
fontLevel = (metric, zone, guide)
glyphLevel = (margin, anchor, fill, oncurve, offcurve, coordinates, redarraow, layer, guide)


class GlyphShowToolbar(BaseToolbar):
    itemType = wx.ITEM_CHECK

    def __init__(self, parent):
        super().__init__(parent, "GlyphShowToolbar")
        self.appendTool(
            "Show Margins",
            "MARGIN",
            "Show glyph margins in current glyph editor",
            margin,
        )
        self.appendTool(
            "Show Metric",
            "METRIC",
            "Show vertical font metric in current glyph editor",
            metric,
        )
        self.appendTool(
            "Show Zones", "ZONES", "Show alignment zones in current glyph editor", zone
        )
        self.appendTool(
            "Show Guides", "GUIDE", "Show guides in current glyph editor", guide
        )
        self.appendTool(
            "Show Anchors", "ANCHOR", "Show anchors in current glyph editor", anchor
        )
        self.tool_oncurve = self.appendTool(
            "Show On-Curve Points",
            "ONCURVE",
            "Show outline fill in current glyph editor",
            oncurve,
        )
        self.tool_offcurve = self.appendTool(
            "Show Off-Curve Points",
            "OFFCURVE",
            "Show outline fill in current glyph editor",
            offcurve,
        )
        self.tool_coordinates = self.appendTool(
            "Show Coordinates",
            "COORDINATES",
            "Show coordinates in current glyph editor",
            coordinates,
        )
        self.appendTool(
            "Show Red Arrow",
            "REDARROW",
            "Show red arraow hints in current glyph editor",
            redarraow,
        )
        self.appendTool(
            "Show Fill", "FILL", "Show outline fill in current glyph editor", fill
        )
        self.AddSeparator()
        tool = self.appendTool(
            "Show Layer",
            "LAYER_VIEW",
            "Show layer in current glyph editor",
            layer,
            wx.ITEM_NORMAL,
        )
        self.SetToolDropDown(tool.Id, True)
        self.Bind(aui.EVT_AUITOOLBAR_TOOL_DROPDOWN, self.on_Tool, tool)

    def on_menu(self, event):
        item = event.EventObject.FindItemById(event.Id)
        planeStack = self.currentView.frame.canvas.getDrawingPlaneStack(
            item.ItemLabelText
        )
        if planeStack:
            planeStack.visibleInactive = event.IsChecked()

    def on_Tool(self, event):
        tool = self.FindTool(event.Id)
        checked = event.IsChecked()
        i = tool.GetUserData()
        canvas = self.currentView.frame.canvas
        if i == layer:
            self.SetToolSticky(event.Id, True)
            menu = wx.Menu()
            for drawingPlane in canvas.layerPlanes:
                mnuItem = menu.Append(
                    wx.ID_ANY,
                    drawingPlane.name,
                    f"Show {drawingPlane.name} in current glyph editor",
                    wx.ITEM_CHECK,
                )
                mnuItem.Check(drawingPlane.visible)
                mnuItem.Enable(not drawingPlane.active)
                self.Bind(wx.EVT_MENU, self.on_menu, mnuItem)
            self.PopupMenu(menu)
            menu.Destroy()
            self.SetToolSticky(event.Id, False)
            event.Skip()
            return
        planes = []
        if i in fontLevel:
            fontPlanes = canvas.getDrawingPlaneStack("Font")
            if i == metric:
                planes.append(fontPlanes.get("VerticalMetric"))
            elif i == zone:
                planes.append(fontPlanes.get("AlignmentZones"))
            elif i == guide:
                planes.append(fontPlanes.get("FontGuide"))
            for plane in planes:
                plane.visible = event.IsChecked()
        planes = []
        activePlanes = canvas.getActiveDrawingPlaneStack()
        if i == margin:
            planes.append(activePlanes.get("GlyphMetric"))
        elif i == guide:
            planes.append(activePlanes.get("GlyphGuide"))
        elif i == anchor:
            planes.append(activePlanes.get("Anchors"))
            planes.append(activePlanes.get("AnchorLabels"))
        elif i == oncurve:
            planes.append(activePlanes.get("OnCurvePoints"))
            if (
                not checked
                or self.tool_coordinates.State & aui.AUI_BUTTON_STATE_CHECKED
            ):
                planes.append(activePlanes.get("OnCurvePointLabels"))
        elif i == offcurve:
            planes.append(activePlanes.get("OffCurvePoints"))
            if (
                not checked
                or self.tool_coordinates.State & aui.AUI_BUTTON_STATE_CHECKED
            ):
                planes.append(activePlanes.get("OffCurvePointLabels"))
        elif i == coordinates:
            if activePlanes.get("OnCurvePoints").visibleActive:
                planes.append(activePlanes.get("OnCurvePointLabels"))
            if activePlanes.get("OffCurvePoints").visibleActive:
                planes.append(activePlanes.get("OffCurvePointLabels"))
        elif i == redarraow:
            planes.append(activePlanes.get("RedArrows"))
        elif i == fill:
            planes.append(activePlanes.get("GlyphFill"))
        for plane in planes:
            plane.visibleActive = checked
        canvas.SetFocus()

    def on_update_Tool(self, event):
        view = self.currentView
        if isinstance(view, UfoGlyphView):
            canvas = view.frame.canvas
            tool = self.FindTool(event.Id)
            i = tool.GetUserData()
            if i == layer:
                event.Enable(len(canvas.layerPlanes) > 1)
                return
            plane = None
            if i in fontLevel:
                event.Enable(True)
                fontPlanes = canvas.getDrawingPlaneStack("Font")
                if i == metric:
                    plane = fontPlanes.get("VerticalMetric")
                elif i == zone:
                    plane = fontPlanes.get("AlignmentZones")
                elif i == guide:
                    plane = fontPlanes.get("FontGuide")
                if plane:
                    event.Check(plane.visible)
            plane = None
            activePlanes = canvas.getActiveDrawingPlaneStack()
            if i == margin:
                plane = activePlanes.get("GlyphMetric")
            elif i == guide:
                plane = activePlanes.get("GlyphGuide")
            elif i == anchor:
                plane = activePlanes.get("Anchors")
            elif i == oncurve:
                plane = activePlanes.get("OnCurvePoints")
            elif i == offcurve:
                plane = activePlanes.get("OffCurvePoints")
            elif i == redarraow:
                plane = activePlanes.get("RedArrows")
            elif i == fill:
                plane = activePlanes.get("GlyphFill")
            if plane:
                event.Check(plane.visibleActive)
            elif i == coordinates:
                event.Check(
                    activePlanes.get("OnCurvePointLabels").visibleActive
                    or activePlanes.get("OffCurvePointLabels").visibleActive
                )
            event.Enable(True)
            return
        event.Enable(False)



"""
toolzoom
===============================================================================
"""

import wx

from wbBase.tools import get_wxBrush, get_wxPen
from ..cursor import cursors

from .base import RUBBERBANDING, GlyphTool


class ToolZoom(GlyphTool):
    cursor = cursors["ZoomIn"]
    brush = get_wxBrush(wx.Colour(200, 200, 200, 64))
    pen = get_wxPen(wx.Colour(128, 128, 128, 128))

    def on_LEFT_DOWN(self, event):
        GlyphTool.on_LEFT_DOWN(self, event)
        self.state = RUBBERBANDING
        event.Skip()

    def on_LEFT_UP(self, event):
        if all([i is not None for i in (self.xClick, self.yClick, self.x, self.y)]):
            x = min(self.xClick, self.x)
            y = min(self.yClick, self.y)
            w = abs(self.xClick - self.x)
            h = abs(self.yClick - self.y)
            if w or h:
                canvas_x = self.canvas.screenToCanvasX(x + w / 2)
                canvas_y = self.canvas.screenToCanvasY(y + h / 2)
                rubberbandRatio = float(w) / float(h)
                screenRatio = float(self.canvas.ClientSize.width) / float(
                    self.canvas.ClientSize.height
                )
                if rubberbandRatio > screenRatio:
                    factor = float(self.canvas.ClientSize.width) / float(w)
                else:
                    factor = float(self.canvas.ClientSize.height) / float(h)
                self.canvas.zoom *= factor
                self.canvas.centerCanvasOnScreen(canvas_x, canvas_y)
        GlyphTool.on_LEFT_UP(self, event)

    def on_MOTION(self, event):
        GlyphTool.on_MOTION(self, event)
        if event.Dragging():
            if self.state == RUBBERBANDING:
                x = min(self.xClick, self.xPrev)
                y = min(self.yClick, self.yPrev)
                w = abs(self.xClick - self.xPrev)
                h = abs(self.yClick - self.yPrev)
                rubberband = wx.Rect(x, y, w, h)
                x = min(self.xClick, self.x)
                y = min(self.yClick, self.y)
                w = abs(self.xClick - self.x)
                h = abs(self.yClick - self.y)
                rubberband.Union(wx.Rect(x, y, w, h))
                rubberband.Inflate(1, 1)
                self.canvas.RefreshRect(rubberband)
        event.Skip()

    def draw(self, gc):
        if self.state == RUBBERBANDING and not None in (
            self.xClick,
            self.yClick,
            self.x,
            self.y,
        ):
            gc.SetPen(self.pen)
            gc.SetBrush(self.brush)
            gc.DrawRectangle(
                min(self.xClick, self.x),
                min(self.yClick, self.y),
                abs(self.xClick - self.x),
                abs(self.yClick - self.y),
            )
